# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Simple DAG with Short Circuit Operator.

It checks:
    - if events that should be emitted are there
    - if events for skipped tasks are not emitted
"""

from __future__ import annotations

from datetime import datetime

from airflow import DAG
from airflow.models import Variable
from airflow.providers.standard.operators.python import PythonOperator, ShortCircuitOperator

from system.openlineage.expected_events import get_expected_event_file_path
from system.openlineage.operator import OpenLineageTestOperator


def do_nothing():
    pass


def check_events_number_func():
    try:
        from airflow.sdk.exceptions import AirflowRuntimeError as ExpectedError  # AF3
    except ImportError:
        ExpectedError = KeyError  # AF2

    for event_type in ("start", "complete"):
        try:
            events = Variable.get(
                key=f"openlineage_short_circuit_dag.should_be_skipped.event.{event_type}",
                deserialize_json=True,
            )
        except ExpectedError:
            pass
        else:
            raise ValueError(
                f"Expected no {event_type.upper()} events for task `should_be_skipped`, got {events}"
            )


DAG_ID = "openlineage_short_circuit_dag"

with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule=None,
    catchup=False,
    default_args={"retries": 0},
) as dag:
    do_nothing_task = PythonOperator(task_id="do_nothing_task", python_callable=do_nothing)

    skip_tasks = ShortCircuitOperator(
        task_id="skip_tasks", python_callable=lambda: False, ignore_downstream_trigger_rules=False
    )

    should_be_skipped = PythonOperator(task_id="should_be_skipped", python_callable=do_nothing)

    check_events_number = PythonOperator(
        task_id="check_events_number", python_callable=check_events_number_func, trigger_rule="none_failed"
    )

    check_events = OpenLineageTestOperator(
        task_id="check_events", file_path=get_expected_event_file_path(DAG_ID)
    )

    do_nothing_task >> skip_tasks >> should_be_skipped >> check_events_number >> check_events


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Complex DAG without schedule, with multiple operators, task groups, dependencies etc.

It checks:
    - required keys
    - field formats and types
    - number of task events (one start, one complete)
    - if EmptyOperator will emit OL events with callback or outlet
    - if EmptyOperator without modification will not emit OL events
    - if CustomOperator without Extractor will emit OL events
    - task groups serialization without dependencies
    - additional task configuration attrs (owner, max_active_tis_per_dag etc.)
"""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any

from airflow import DAG
from airflow.models import Variable
from airflow.providers.common.compat.assets import Asset
from airflow.providers.common.compat.sdk import BaseOperator
from airflow.providers.standard.operators.bash import BashOperator
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.providers.standard.operators.python import PythonOperator

try:
    from airflow.sdk import TaskGroup
except ImportError:
    from airflow.utils.task_group import TaskGroup  # type: ignore[no-redef]

from system.openlineage.expected_events import AIRFLOW_VERSION, get_expected_event_file_path
from system.openlineage.operator import OpenLineageTestOperator


def check_events_number_func():
    for event_type in ("start", "complete", "fail"):
        try:
            Variable.get(key=f"openlineage_base_complex_dag.task_0.event.{event_type}", deserialize_json=True)
        except Exception:
            pass
        else:
            raise ValueError("Expected no events for task `task_0`.")


def do_nothing():
    pass


class SomeCustomOperator(BashOperator):
    def __init__(self, **kwargs):
        # Just to test that these attrs are included in OL event
        self.deferrable = True
        self.external_dag_id = "external_dag_id"
        self.external_task_id = "external_task_id"
        super().__init__(**kwargs)


class CustomMappedOperator(BaseOperator):
    def __init__(self, value, **kwargs):
        super().__init__(**kwargs)
        self.value = value

    def execute(self, context):
        return self.value + 1


DAG_ID = "openlineage_base_complex_dag"

with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule=None,
    catchup=False,
    description="OpenLineage complex DAG description",
    owner_links={"airflow": "https://airflow.apache.org/"},
    tags=["first", "second@", "with'quote", 'z"e'],
    default_args={"retries": 0},
) as dag:
    # task_0 will not emit any events, but the owner will be picked up and added to DAG
    task_0 = EmptyOperator(task_id="task_0", owner='owner"1')
    task_1 = BashOperator(
        task_id="task_1.id.with.dots",
        bash_command="exit 0;",
        owner="owner'2",
        execution_timeout=timedelta(seconds=456),
        doc_rst="RST doc",
    )
    task_2 = PythonOperator(
        task_id="task_2",
        python_callable=do_nothing,
        inlets=[Asset(uri="s3://bucket2/dir2/file2.txt"), Asset(uri="s3://bucket2/dir2/file3.txt")],
        max_retry_delay=42,
        doc="text doc",
        doc_md="should be skipped",
        doc_json="should be skipped",
        doc_yaml="should be skipped",
        doc_rst="should be skipped",
    )
    task_3 = EmptyOperator(
        task_id="task_3",
        outlets=[Asset(uri="s3://bucket/dir/file.txt")],
        doc_md="MD doc",
        doc_json="should be skipped",
        doc_yaml="should be skipped",
        doc_rst="should be skipped",
    )
    task_4 = SomeCustomOperator(
        task_id="task_4",
        bash_command="exit 0;",
        owner="owner3",
        max_active_tis_per_dag=7,
        max_active_tis_per_dagrun=2,
        doc_json="JSON doc",
        doc_yaml="should be skipped",
        doc_rst="should be skipped",
    )

    with TaskGroup("section_1", prefix_group_id=True) as tg:
        task_5 = CustomMappedOperator.partial(task_id="task_5", doc_md="md doc").expand(value=[1])
        with TaskGroup("section_2", parent_group=tg, tooltip="group_tooltip") as tg2:
            add_args: dict[str, Any] = {"sla": timedelta(seconds=123)} if AIRFLOW_VERSION.major == 2 else {}
            task_6 = EmptyOperator(
                task_id="task_6",
                on_success_callback=lambda x: print(1),
                doc_yaml="YAML doc",
                doc_rst="should be skipped",
                **add_args,
            )
            with TaskGroup("section_3", parent_group=tg2):
                task_7 = PythonOperator(task_id="task_7", python_callable=lambda: 1)

    check_events_number = PythonOperator(
        task_id="check_events_number", python_callable=check_events_number_func
    )

    check_events = OpenLineageTestOperator(
        task_id="check_events", file_path=get_expected_event_file_path(DAG_ID)
    )

    task_1 >> [task_2, task_7] >> check_events_number
    task_2 >> task_3 >> [task_4, task_5] >> task_6 >> check_events_number
    check_events_number >> check_events


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from attrs import Factory, define
from openlineage.client.event_v2 import Dataset as OLDataset

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    from openlineage.client.facet import BaseFacet as BaseFacet_V1
from openlineage.client.facet_v2 import JobFacet, RunFacet

from airflow.utils.log.logging_mixin import LoggingMixin

# this is not to break static checks compatibility with v1 OpenLineage facet classes
DatasetSubclass = TypeVar("DatasetSubclass", bound=OLDataset)
BaseFacetSubclass = TypeVar("BaseFacetSubclass", bound=BaseFacet_V1 | RunFacet | JobFacet)

OL_METHOD_NAME_START = "get_openlineage_facets_on_start"
OL_METHOD_NAME_COMPLETE = "get_openlineage_facets_on_complete"
OL_METHOD_NAME_FAIL = "get_openlineage_facets_on_failure"


@define
class OperatorLineage(Generic[DatasetSubclass, BaseFacetSubclass]):
    """Structure returned from lineage extraction."""

    inputs: list[DatasetSubclass] = Factory(list)
    outputs: list[DatasetSubclass] = Factory(list)
    run_facets: dict[str, BaseFacetSubclass] = Factory(dict)
    job_facets: dict[str, BaseFacetSubclass] = Factory(dict)


class BaseExtractor(ABC, LoggingMixin):
    """
    Abstract base extractor class.

    This is used mostly to maintain support for custom extractors.
    """

    _allowed_query_params: list[str] = []

    def __init__(self, operator):
        super().__init__()
        self.operator = operator

    @classmethod
    @abstractmethod
    def get_operator_classnames(cls) -> list[str]:
        """
        Get a list of operators that extractor works for.

        This is an abstract method that subclasses should implement. There are
        operators that work very similarly and one extractor can cover.
        """
        raise NotImplementedError()

    @abstractmethod
    def _execute_extraction(self) -> OperatorLineage | None: ...

    def extract(self) -> OperatorLineage | None:
        return self._execute_extraction()

    def extract_on_complete(self, task_instance) -> OperatorLineage | None:
        return self.extract()

    def extract_on_failure(self, task_instance) -> OperatorLineage | None:
        return self.extract_on_complete(task_instance)


class DefaultExtractor(BaseExtractor):
    """Extractor that uses `get_openlineage_facets_on_start/complete/failure` methods."""

    @classmethod
    def get_operator_classnames(cls) -> list[str]:
        """
        Assign this extractor to *no* operators.

        Default extractor is chosen not on the classname basis, but
        by existence of get_openlineage_facets method on operator.
        """
        return []

    def _execute_extraction(self) -> OperatorLineage | None:
        method = getattr(self.operator, OL_METHOD_NAME_START, None)
        if callable(method):
            self.log.debug(
                "Trying to execute '%s' method of '%s'.", OL_METHOD_NAME_START, self.operator.task_type
            )
            return self._get_openlineage_facets(method)
        self.log.debug(
            "Operator '%s' does not have '%s' method.", self.operator.task_type, OL_METHOD_NAME_START
        )
        return OperatorLineage()

    def extract_on_complete(self, task_instance) -> OperatorLineage | None:
        method = getattr(self.operator, OL_METHOD_NAME_COMPLETE, None)
        if callable(method):
            self.log.debug(
                "Trying to execute '%s' method of '%s'.", OL_METHOD_NAME_COMPLETE, self.operator.task_type
            )
            return self._get_openlineage_facets(method, task_instance)
        self.log.debug(
            "Operator '%s' does not have '%s' method.", self.operator.task_type, OL_METHOD_NAME_COMPLETE
        )
        return self.extract()

    def extract_on_failure(self, task_instance) -> OperatorLineage | None:
        method = getattr(self.operator, OL_METHOD_NAME_FAIL, None)
        if callable(method):
            self.log.debug(
                "Trying to execute '%s' method of '%s'.", OL_METHOD_NAME_FAIL, self.operator.task_type
            )
            return self._get_openlineage_facets(method, task_instance)
        self.log.debug(
            "Operator '%s' does not have '%s' method.", self.operator.task_type, OL_METHOD_NAME_FAIL
        )
        return self.extract_on_complete(task_instance)

    def _get_openlineage_facets(self, get_facets_method, *args) -> OperatorLineage | None:
        try:
            facets: OperatorLineage | None = get_facets_method(*args)
            if facets is None:
                self.log.debug("OpenLineage method returned `None`")
                return None
            # "rewrite" OperatorLineage to safeguard against different version of the same class
            # that was existing in openlineage-airflow package outside of Airflow repo
            return OperatorLineage(
                inputs=facets.inputs,
                outputs=facets.outputs,
                run_facets=facets.run_facets,
                job_facets=facets.job_facets,
            )
        except ImportError:
            self.log.exception(
                "OpenLineage provider method failed to import OpenLineage integration. "
                "This should not happen."
            )
        except Exception as e:
            self.log.warning(
                "OpenLineage method failed to extract data from Operator with the following exception: `%s`",
                e,
            )
            self.log.debug("OpenLineage extraction failure details:", exc_info=True)
        return None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from airflow.providers.openlineage.plugins.listener import get_openlineage_listener
from airflow.providers.openlineage.plugins.macros import (
    lineage_job_name,
    lineage_job_namespace,
    lineage_root_job_name,
    lineage_root_run_id,
    lineage_run_id,
)

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context

log = logging.getLogger(__name__)


def _get_parent_job_information_as_spark_properties(context: Context) -> dict:
    """
    Retrieve parent job information as Spark properties.

    Args:
        context: The context containing task instance information.

    Returns:
        Spark properties with the parent job information.
    """
    ti = context["ti"]
    return {
        "spark.openlineage.parentJobNamespace": lineage_job_namespace(),
        "spark.openlineage.parentJobName": lineage_job_name(ti),  # type: ignore[arg-type]
        "spark.openlineage.parentRunId": lineage_run_id(ti),  # type: ignore[arg-type]
        "spark.openlineage.rootParentRunId": lineage_root_run_id(ti),  # type: ignore[arg-type]
        "spark.openlineage.rootParentJobName": lineage_root_job_name(ti),  # type: ignore[arg-type]
        "spark.openlineage.rootParentJobNamespace": lineage_job_namespace(),
    }


def _get_transport_information_as_spark_properties() -> dict:
    """Retrieve transport information as Spark properties."""

    def _get_transport_information(tp) -> dict:
        properties = {
            "type": tp.kind,
            "url": tp.url,
            "endpoint": tp.endpoint,
            "timeoutInMillis": str(
                int(tp.timeout) * 1000  # convert to milliseconds, as required by Spark integration
            ),
        }
        if hasattr(tp, "compression") and tp.compression:
            properties["compression"] = str(tp.compression)

        if hasattr(tp.config.auth, "api_key") and tp.config.auth.get_bearer():
            properties["auth.type"] = "api_key"
            properties["auth.apiKey"] = tp.config.auth.get_bearer()

        if hasattr(tp.config, "custom_headers") and tp.config.custom_headers:
            for key, value in tp.config.custom_headers.items():
                properties[f"headers.{key}"] = value
        return properties

    def _format_transport(props: dict, transport: dict, name: str | None):
        for key, value in transport.items():
            if name:
                props[f"spark.openlineage.transport.transports.{name}.{key}"] = value
            else:
                props[f"spark.openlineage.transport.{key}"] = value
        return props

    transport = get_openlineage_listener().adapter.get_or_create_openlineage_client().transport

    if transport.kind == "composite":
        http_transports = {}
        for nested_transport in transport.transports:
            if nested_transport.kind == "http":
                http_transports[nested_transport.name] = _get_transport_information(nested_transport)
            else:
                name = nested_transport.name if hasattr(nested_transport, "name") else "no-name"
                log.info(
                    "OpenLineage transport type `%s` with name `%s` is not supported in composite transport.",
                    nested_transport.kind,
                    name,
                )
        if len(http_transports) == 0:
            log.warning(
                "OpenLineage transport type `composite` does not contain http transport. Skipping "
                "injection of OpenLineage transport information into Spark properties.",
            )
            return {}
        props = {
            "spark.openlineage.transport.type": "composite",
            "spark.openlineage.transport.continueOnFailure": str(transport.config.continue_on_failure),
        }
        for name, http_transport in http_transports.items():
            props = _format_transport(props, http_transport, name)
        return props

    if transport.kind == "http":
        return _format_transport({}, _get_transport_information(transport), None)

    log.info(
        "OpenLineage transport type `%s` does not support automatic "
        "injection of OpenLineage transport information into Spark properties.",
        transport.kind,
    )
    return {}


def _is_parent_job_information_present_in_spark_properties(properties: dict) -> bool:
    """
    Check if any parent job information is present in Spark properties.

    Args:
        properties: Spark properties.

    Returns:
        True if parent job information is present, False otherwise.
    """
    return any(str(key).startswith("spark.openlineage.parent") for key in properties)


def _is_transport_information_present_in_spark_properties(properties: dict) -> bool:
    """
    Check if any transport information is present in Spark properties.

    Args:
        properties: Spark properties.

    Returns:
        True if transport information is present, False otherwise.
    """
    return any(str(key).startswith("spark.openlineage.transport") for key in properties)


def inject_parent_job_information_into_spark_properties(properties: dict, context: Context) -> dict:
    """
    Inject parent job information into Spark properties if not already present.

    Args:
        properties: Spark properties.
        context: The context containing task instance information.

    Returns:
        Modified Spark properties with OpenLineage parent job information properties injected, if applicable.
    """
    if _is_parent_job_information_present_in_spark_properties(properties):
        log.info(
            "Some OpenLineage properties with parent job information are already present "
            "in Spark properties. Skipping the injection of OpenLineage "
            "parent job information into Spark properties."
        )
        return properties

    return {**properties, **_get_parent_job_information_as_spark_properties(context)}


def inject_transport_information_into_spark_properties(properties: dict, context: Context) -> dict:
    """
    Inject transport information into Spark properties if not already present.

    Args:
        properties: Spark properties.
        context: The context containing task instance information.

    Returns:
        Modified Spark properties with OpenLineage transport information properties injected, if applicable.
    """
    if _is_transport_information_present_in_spark_properties(properties):
        log.info(
            "Some OpenLineage properties with transport information are already present "
            "in Spark properties. Skipping the injection of OpenLineage "
            "transport information into Spark properties."
        )
        return properties

    return {**properties, **_get_transport_information_as_spark_properties()}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import os
import pathlib
import uuid
from unittest import mock
from unittest.mock import ANY, MagicMock, call, patch

import pytest
from openlineage.client.event_v2 import Dataset, Job, Run, RunEvent, RunState
from openlineage.client.facet_v2 import (
    documentation_job,
    error_message_run,
    external_query_run,
    job_type_job,
    nominal_time_run,
    ownership_job,
    parent_run,
    processing_engine_run,
    sql_job,
    tags_job,
)

from airflow import DAG
from airflow.models.dagrun import DagRun, DagRunState
from airflow.models.taskinstance import TaskInstance, TaskInstanceState
from airflow.providers.openlineage.conf import namespace
from airflow.providers.openlineage.extractors import OperatorLineage
from airflow.providers.openlineage.plugins.adapter import _PRODUCER, OpenLineageAdapter
from airflow.providers.openlineage.plugins.facets import (
    AirflowDagRunFacet,
    AirflowDebugRunFacet,
    AirflowStateRunFacet,
)
from airflow.providers.openlineage.utils.utils import get_airflow_job_facet
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.utils.task_group import TaskGroup
from airflow.utils.types import DagRunType

from tests_common.test_utils.compat import BashOperator
from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.markers import skip_if_force_lowest_dependencies_marker
from tests_common.test_utils.taskinstance import create_task_instance
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS


@pytest.mark.parametrize(
    ("env_vars", "expected_logging"),
    [
        ({"AIRFLOW__LOGGING__LOGGING_LEVEL": "DEBUG"}, "DEBUG"),
        ({"AIRFLOW__LOGGING__LOGGING_LEVEL": "INFO"}, None),
        ({}, None),  # When no value is provided, default should be INFO and propagation is skipped.
    ],
)
def test_create_client_logging_propagation(env_vars, expected_logging):
    with patch.dict(os.environ, env_vars, clear=True):
        assert os.getenv("OPENLINEAGE_CLIENT_LOGGING") is None
        OpenLineageAdapter().get_or_create_openlineage_client()
        assert os.getenv("OPENLINEAGE_CLIENT_LOGGING") == expected_logging


@patch.dict(
    os.environ,
    {"OPENLINEAGE_URL": "http://ol-api:5000", "OPENLINEAGE_API_KEY": "api-key"},
)
def test_create_client_from_ol_env():
    client = OpenLineageAdapter().get_or_create_openlineage_client()

    assert client.transport.url == "http://ol-api:5000"


@conf_vars(
    {
        ("openlineage", "transport"): '{"type": "http", "url": "http://ol-api:5000",'
        ' "auth": {"type": "api_key", "apiKey": "api-key"}}'
    }
)
def test_create_client_from_config_with_options():
    client = OpenLineageAdapter().get_or_create_openlineage_client()

    assert client.transport.kind == "http"
    assert client.transport.url == "http://ol-api:5000"


def test_create_client_from_yaml_config():
    current_folder = pathlib.Path(__file__).parent.resolve()
    yaml_config = str((current_folder / "openlineage_configs" / "http.yaml").resolve())
    with conf_vars({("openlineage", "config_path"): yaml_config}):
        client = OpenLineageAdapter().get_or_create_openlineage_client()

    assert client.transport.kind == "http"


def test_create_client_from_env_var_config():
    current_folder = pathlib.Path(__file__).parent.resolve()
    yaml_config = str((current_folder / "openlineage_configs" / "http.yaml").resolve())
    with patch.dict(os.environ, {"OPENLINEAGE_CONFIG": yaml_config}):
        client = OpenLineageAdapter().get_or_create_openlineage_client()

    assert client.transport.kind == "http"
    assert client.transport.url == "http://localhost:5050"


@patch.dict(
    os.environ,
    {
        "OPENLINEAGE_URL": "http://ol-from-env:5000",
        "OPENLINEAGE_API_KEY": "api-key-from-env",
    },
)
@patch.dict(os.environ, {"OPENLINEAGE_CONFIG": "some/config.yml"})
def test_create_client_overrides_env_vars():
    current_folder = pathlib.Path(__file__).parent.resolve()
    yaml_config = str((current_folder / "openlineage_configs" / "http.yaml").resolve())
    with conf_vars({("openlineage", "config_path"): yaml_config}):
        client = OpenLineageAdapter().get_or_create_openlineage_client()

        assert client.transport.kind == "http"
        assert client.transport.url == "http://localhost:5050"

    with conf_vars({("openlineage", "transport"): '{"type": "console"}'}):
        client = OpenLineageAdapter().get_or_create_openlineage_client()

        assert client.transport.kind == "console"


@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.timer")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.incr")
def test_emit_start_event(mock_stats_incr, mock_stats_timer):
    client = MagicMock()
    adapter = OpenLineageAdapter(client)

    run_id = str(uuid.uuid4())
    event_time = datetime.datetime.now().isoformat()
    adapter.start_task(
        run_id=run_id,
        job_name="job",
        job_description="description",
        event_time=event_time,
        nominal_start_time=datetime.datetime(2022, 1, 1).isoformat(),
        nominal_end_time=datetime.datetime(2022, 1, 1).isoformat(),
        owners=[],
        tags=[],
        task=None,
        run_facets=None,
    )

    assert (
        call(
            RunEvent(
                eventType=RunState.START,
                eventTime=event_time,
                run=Run(
                    runId=run_id,
                    facets={
                        "nominalTime": nominal_time_run.NominalTimeRunFacet(
                            nominalStartTime="2022-01-01T00:00:00",
                            nominalEndTime="2022-01-01T00:00:00",
                        ),
                        "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                            version=ANY, name="Airflow", openlineageAdapterVersion=ANY
                        ),
                    },
                ),
                job=Job(
                    namespace=namespace(),
                    name="job",
                    facets={
                        "documentation": documentation_job.DocumentationJobFacet(description="description"),
                        "jobType": job_type_job.JobTypeJobFacet(
                            processingType="BATCH", integration="AIRFLOW", jobType="TASK"
                        ),
                    },
                ),
                producer=_PRODUCER,
                inputs=[],
                outputs=[],
            )
        )
        in client.emit.mock_calls
    )

    mock_stats_incr.assert_not_called()
    mock_stats_timer.assert_called_with("ol.emit.attempts")


@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.timer")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.incr")
def test_emit_start_event_with_additional_information(mock_stats_incr, mock_stats_timer):
    client = MagicMock()
    adapter = OpenLineageAdapter(client)

    run_id = str(uuid.uuid4())
    parent_run_id = str(uuid.uuid4())
    event_time = datetime.datetime.now().isoformat()
    adapter.start_task(
        run_id=run_id,
        job_name="job",
        job_description="description",
        job_description_type="text/plain",
        event_time=event_time,
        nominal_start_time=datetime.datetime(2022, 1, 1).isoformat(),
        nominal_end_time=datetime.datetime(2022, 1, 1).isoformat(),
        owners=["owner1", "owner2"],
        tags=["tag1", "tag2"],
        task=OperatorLineage(
            inputs=[Dataset(namespace="bigquery", name="a.b.c"), Dataset(namespace="bigquery", name="x.y.z")],
            outputs=[Dataset(namespace="gs://bucket", name="exported_folder")],
            job_facets={"sql": sql_job.SQLJobFacet(query="SELECT 1;")},
            run_facets={
                "parent": parent_run.ParentRunFacet(
                    run=parent_run.Run(runId=parent_run_id),
                    job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                    root=parent_run.Root(
                        run=parent_run.RootRun(runId=parent_run_id),
                        job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                    ),
                ),
                "externalQuery1": external_query_run.ExternalQueryRunFacet(
                    externalQueryId="123", source="source"
                ),
            },
        ),
        run_facets={
            "externalQuery2": external_query_run.ExternalQueryRunFacet(externalQueryId="999", source="source")
        },
    )

    assert (
        call(
            RunEvent(
                eventType=RunState.START,
                eventTime=event_time,
                run=Run(
                    runId=run_id,
                    facets={
                        "nominalTime": nominal_time_run.NominalTimeRunFacet(
                            nominalStartTime="2022-01-01T00:00:00",
                            nominalEndTime="2022-01-01T00:00:00",
                        ),
                        "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                            version=ANY, name="Airflow", openlineageAdapterVersion=ANY
                        ),
                        "parent": parent_run.ParentRunFacet(
                            run=parent_run.Run(runId=parent_run_id),
                            job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                            root=parent_run.Root(
                                run=parent_run.RootRun(runId=parent_run_id),
                                job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                            ),
                        ),
                        "externalQuery1": external_query_run.ExternalQueryRunFacet(
                            externalQueryId="123", source="source"
                        ),
                        "externalQuery2": external_query_run.ExternalQueryRunFacet(
                            externalQueryId="999", source="source"
                        ),
                    },
                ),
                job=Job(
                    namespace=namespace(),
                    name="job",
                    facets={
                        "documentation": documentation_job.DocumentationJobFacet(
                            description="description", contentType="text/plain"
                        ),
                        "ownership": ownership_job.OwnershipJobFacet(
                            owners=[
                                ownership_job.Owner(name="owner1", type=None),
                                ownership_job.Owner(name="owner2", type=None),
                            ]
                        ),
                        "sql": sql_job.SQLJobFacet(query="SELECT 1;"),
                        "jobType": job_type_job.JobTypeJobFacet(
                            processingType="BATCH", integration="AIRFLOW", jobType="TASK"
                        ),
                        "tags": tags_job.TagsJobFacet(
                            tags=[
                                tags_job.TagsJobFacetFields(key="tag1", value="tag1", source="AIRFLOW"),
                                tags_job.TagsJobFacetFields(key="tag2", value="tag2", source="AIRFLOW"),
                            ]
                        ),
                    },
                ),
                producer=_PRODUCER,
                inputs=[
                    Dataset(namespace="bigquery", name="a.b.c"),
                    Dataset(namespace="bigquery", name="x.y.z"),
                ],
                outputs=[Dataset(namespace="gs://bucket", name="exported_folder")],
            )
        )
        in client.emit.mock_calls
    )

    mock_stats_incr.assert_not_called()
    mock_stats_timer.assert_called_with("ol.emit.attempts")


@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.timer")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.incr")
def test_emit_complete_event(mock_stats_incr, mock_stats_timer):
    client = MagicMock()
    adapter = OpenLineageAdapter(client)

    run_id = str(uuid.uuid4())
    event_time = datetime.datetime.now().isoformat()
    adapter.complete_task(
        run_id=run_id,
        end_time=event_time,
        job_name="job",
        task=OperatorLineage(),
        owners=[],
        tags=[],
        job_description=None,
        nominal_start_time=datetime.datetime(2022, 1, 1).isoformat(),
        nominal_end_time=datetime.datetime(2022, 1, 1).isoformat(),
    )

    assert (
        call(
            RunEvent(
                eventType=RunState.COMPLETE,
                eventTime=event_time,
                run=Run(
                    runId=run_id,
                    facets={
                        "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                            version=ANY, name="Airflow", openlineageAdapterVersion=ANY
                        ),
                        "nominalTime": nominal_time_run.NominalTimeRunFacet(
                            nominalStartTime="2022-01-01T00:00:00",
                            nominalEndTime="2022-01-01T00:00:00",
                        ),
                    },
                ),
                job=Job(
                    namespace=namespace(),
                    name="job",
                    facets={
                        "jobType": job_type_job.JobTypeJobFacet(
                            processingType="BATCH", integration="AIRFLOW", jobType="TASK"
                        )
                    },
                ),
                producer=_PRODUCER,
                inputs=[],
                outputs=[],
            )
        )
        in client.emit.mock_calls
    )

    mock_stats_incr.assert_not_called()
    mock_stats_timer.assert_called_with("ol.emit.attempts")


@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.timer")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.incr")
def test_emit_complete_event_with_additional_information(mock_stats_incr, mock_stats_timer):
    client = MagicMock()
    adapter = OpenLineageAdapter(client)

    run_id = str(uuid.uuid4())
    parent_run_id = str(uuid.uuid4())
    event_time = datetime.datetime.now().isoformat()
    adapter.complete_task(
        run_id=run_id,
        end_time=event_time,
        job_name="job",
        owners=["owner1", "owner2"],
        tags=["tag1", "tag2"],
        job_description="description",
        job_description_type="text/plain",
        nominal_start_time=datetime.datetime(2022, 1, 1).isoformat(),
        nominal_end_time=datetime.datetime(2022, 1, 1).isoformat(),
        task=OperatorLineage(
            inputs=[Dataset(namespace="bigquery", name="a.b.c"), Dataset(namespace="bigquery", name="x.y.z")],
            outputs=[Dataset(namespace="gs://bucket", name="exported_folder")],
            job_facets={"sql": sql_job.SQLJobFacet(query="SELECT 1;")},
            run_facets={
                "externalQuery": external_query_run.ExternalQueryRunFacet(
                    externalQueryId="123", source="source"
                )
            },
        ),
        run_facets={
            "parent": parent_run.ParentRunFacet(
                run=parent_run.Run(runId=parent_run_id),
                job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                root=parent_run.Root(
                    run=parent_run.RootRun(runId=parent_run_id),
                    job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                ),
            ),
            "externalQuery2": external_query_run.ExternalQueryRunFacet(
                externalQueryId="999", source="source"
            ),
        },
    )

    assert (
        call(
            RunEvent(
                eventType=RunState.COMPLETE,
                eventTime=event_time,
                run=Run(
                    runId=run_id,
                    facets={
                        "parent": parent_run.ParentRunFacet(
                            run=parent_run.Run(runId=parent_run_id),
                            job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                            root=parent_run.Root(
                                run=parent_run.RootRun(runId=parent_run_id),
                                job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                            ),
                        ),
                        "nominalTime": nominal_time_run.NominalTimeRunFacet(
                            nominalStartTime="2022-01-01T00:00:00",
                            nominalEndTime="2022-01-01T00:00:00",
                        ),
                        "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                            version=ANY, name="Airflow", openlineageAdapterVersion=ANY
                        ),
                        "externalQuery": external_query_run.ExternalQueryRunFacet(
                            externalQueryId="123", source="source"
                        ),
                        "externalQuery2": external_query_run.ExternalQueryRunFacet(
                            externalQueryId="999", source="source"
                        ),
                    },
                ),
                job=Job(
                    namespace="default",
                    name="job",
                    facets={
                        "documentation": documentation_job.DocumentationJobFacet(
                            description="description", contentType="text/plain"
                        ),
                        "ownership": ownership_job.OwnershipJobFacet(
                            owners=[
                                ownership_job.Owner(name="owner1", type=None),
                                ownership_job.Owner(name="owner2", type=None),
                            ]
                        ),
                        "sql": sql_job.SQLJobFacet(query="SELECT 1;"),
                        "jobType": job_type_job.JobTypeJobFacet(
                            processingType="BATCH", integration="AIRFLOW", jobType="TASK"
                        ),
                        "tags": tags_job.TagsJobFacet(
                            tags=[
                                tags_job.TagsJobFacetFields(key="tag1", value="tag1", source="AIRFLOW"),
                                tags_job.TagsJobFacetFields(key="tag2", value="tag2", source="AIRFLOW"),
                            ]
                        ),
                    },
                ),
                producer=_PRODUCER,
                inputs=[
                    Dataset(namespace="bigquery", name="a.b.c"),
                    Dataset(namespace="bigquery", name="x.y.z"),
                ],
                outputs=[Dataset(namespace="gs://bucket", name="exported_folder")],
            )
        )
        in client.emit.mock_calls
    )

    mock_stats_incr.assert_not_called()
    mock_stats_timer.assert_called_with("ol.emit.attempts")


@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.timer")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.incr")
def test_emit_failed_event(mock_stats_incr, mock_stats_timer):
    client = MagicMock()
    adapter = OpenLineageAdapter(client)

    run_id = str(uuid.uuid4())
    event_time = datetime.datetime.now().isoformat()
    adapter.fail_task(
        run_id=run_id,
        end_time=event_time,
        job_name="job",
        task=OperatorLineage(),
        owners=[],
        tags=[],
        job_description=None,
        nominal_start_time=datetime.datetime(2022, 1, 1).isoformat(),
        nominal_end_time=datetime.datetime(2022, 1, 1).isoformat(),
    )

    assert (
        call(
            RunEvent(
                eventType=RunState.FAIL,
                eventTime=event_time,
                run=Run(
                    runId=run_id,
                    facets={
                        "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                            version=ANY, name="Airflow", openlineageAdapterVersion=ANY
                        ),
                        "nominalTime": nominal_time_run.NominalTimeRunFacet(
                            nominalStartTime="2022-01-01T00:00:00",
                            nominalEndTime="2022-01-01T00:00:00",
                        ),
                    },
                ),
                job=Job(
                    namespace=namespace(),
                    name="job",
                    facets={
                        "jobType": job_type_job.JobTypeJobFacet(
                            processingType="BATCH", integration="AIRFLOW", jobType="TASK"
                        )
                    },
                ),
                producer=_PRODUCER,
                inputs=[],
                outputs=[],
            )
        )
        in client.emit.mock_calls
    )

    mock_stats_incr.assert_not_called()
    mock_stats_timer.assert_called_with("ol.emit.attempts")


@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.timer")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.incr")
def test_emit_failed_event_with_additional_information(mock_stats_incr, mock_stats_timer):
    client = MagicMock()
    adapter = OpenLineageAdapter(client)

    run_id = str(uuid.uuid4())
    parent_run_id = str(uuid.uuid4())
    event_time = datetime.datetime.now().isoformat()
    adapter.fail_task(
        run_id=run_id,
        end_time=event_time,
        job_name="job",
        owners=["owner1", "owner2"],
        tags=["tag1", "tag2"],
        job_description="description",
        job_description_type="text/plain",
        nominal_start_time=datetime.datetime(2022, 1, 1).isoformat(),
        nominal_end_time=datetime.datetime(2022, 1, 1).isoformat(),
        task=OperatorLineage(
            inputs=[Dataset(namespace="bigquery", name="a.b.c"), Dataset(namespace="bigquery", name="x.y.z")],
            outputs=[Dataset(namespace="gs://bucket", name="exported_folder")],
            run_facets={
                "externalQuery": external_query_run.ExternalQueryRunFacet(
                    externalQueryId="123", source="source"
                )
            },
            job_facets={"sql": sql_job.SQLJobFacet(query="SELECT 1;")},
        ),
        run_facets={
            "parent": parent_run.ParentRunFacet(
                run=parent_run.Run(runId=parent_run_id),
                job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                root=parent_run.Root(
                    run=parent_run.RootRun(runId=parent_run_id),
                    job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                ),
            ),
            "externalQuery2": external_query_run.ExternalQueryRunFacet(
                externalQueryId="999", source="source"
            ),
        },
        error=ValueError("Error message"),
    )

    assert client.emit.mock_calls[0] == call(
        RunEvent(
            eventType=RunState.FAIL,
            eventTime=event_time,
            run=Run(
                runId=run_id,
                facets={
                    "parent": parent_run.ParentRunFacet(
                        run=parent_run.Run(runId=parent_run_id),
                        job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                        root=parent_run.Root(
                            run=parent_run.RootRun(runId=parent_run_id),
                            job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                        ),
                    ),
                    "nominalTime": nominal_time_run.NominalTimeRunFacet(
                        nominalStartTime="2022-01-01T00:00:00",
                        nominalEndTime="2022-01-01T00:00:00",
                    ),
                    "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                        version=ANY, name="Airflow", openlineageAdapterVersion=ANY
                    ),
                    "errorMessage": error_message_run.ErrorMessageRunFacet(
                        message="Error message", programmingLanguage="python", stackTrace=None
                    ),
                    "externalQuery": external_query_run.ExternalQueryRunFacet(
                        externalQueryId="123", source="source"
                    ),
                    "externalQuery2": external_query_run.ExternalQueryRunFacet(
                        externalQueryId="999", source="source"
                    ),
                },
            ),
            job=Job(
                namespace="default",
                name="job",
                facets={
                    "documentation": documentation_job.DocumentationJobFacet(
                        description="description", contentType="text/plain"
                    ),
                    "ownership": ownership_job.OwnershipJobFacet(
                        owners=[
                            ownership_job.Owner(name="owner1", type=None),
                            ownership_job.Owner(name="owner2", type=None),
                        ]
                    ),
                    "sql": sql_job.SQLJobFacet(query="SELECT 1;"),
                    "jobType": job_type_job.JobTypeJobFacet(
                        processingType="BATCH", integration="AIRFLOW", jobType="TASK"
                    ),
                    "tags": tags_job.TagsJobFacet(
                        tags=[
                            tags_job.TagsJobFacetFields(key="tag1", value="tag1", source="AIRFLOW"),
                            tags_job.TagsJobFacetFields(key="tag2", value="tag2", source="AIRFLOW"),
                        ]
                    ),
                },
            ),
            producer=_PRODUCER,
            inputs=[
                Dataset(namespace="bigquery", name="a.b.c"),
                Dataset(namespace="bigquery", name="x.y.z"),
            ],
            outputs=[Dataset(namespace="gs://bucket", name="exported_folder")],
        )
    )

    mock_stats_incr.assert_not_called()
    mock_stats_timer.assert_called_with("ol.emit.attempts")


@mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
@mock.patch("airflow.providers.openlineage.plugins.adapter.generate_static_uuid")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.timer")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.incr")
def test_emit_dag_started_event(mock_stats_incr, mock_stats_timer, generate_static_uuid, mock_debug_mode):
    random_uuid = "9d3b14f7-de91-40b6-aeef-e887e2c7673e"
    client = MagicMock()
    adapter = OpenLineageAdapter(client)
    event_time = datetime.datetime.fromisoformat("2021-01-01T00:00:00+00:00")
    dag_id = "dag_id"
    run_id = str(uuid.uuid4())

    with DAG(
        dag_id=dag_id,
        schedule=datetime.timedelta(days=1),
        start_date=datetime.datetime(2024, 6, 1),
        description="dag desc",
        tags=["mytag1"],
    ) as dag:
        tg = TaskGroup(group_id="tg1")
        tg2 = TaskGroup(group_id="tg2", parent_group=tg)
        task_0 = BashOperator(task_id="task_0", bash_command="exit 0;")  # noqa: F841
        task_1 = BashOperator(task_id="task_1", bash_command="exit 0;", task_group=tg)  # noqa: F841
        task_2 = EmptyOperator(task_id="task_2.test.dot", task_group=tg2)  # noqa: F841

    if AIRFLOW_V_3_0_PLUS:
        dag_run = DagRun(
            dag_id=dag_id,
            run_id=run_id,
            start_date=event_time,
            logical_date=event_time,
            data_interval=(event_time, event_time),
        )
    else:
        dag_run = DagRun(
            dag_id=dag_id,
            run_id=run_id,
            start_date=event_time,
            execution_date=event_time,
            data_interval=(event_time, event_time),
        )
    dag_run.dag = dag
    generate_static_uuid.return_value = random_uuid

    job_facets = {**get_airflow_job_facet(dag_run)}

    expected_dag_info = {
        "timetable": {"delta": 86400.0},
        "dag_id": dag_id,
        "description": "dag desc",
        "owner": "airflow",
        "start_date": "2024-06-01T00:00:00+00:00",
        "tags": "['mytag1']",
        "fileloc": pathlib.Path(__file__).resolve().as_posix(),
    }
    if hasattr(dag, "schedule_interval"):  # Airflow 2 compat.
        expected_dag_info["schedule_interval"] = "86400.0 seconds"
    else:  # Airflow 3 and up.
        expected_dag_info["timetable_summary"] = "1 day, 0:00:00"

    dag_run_facet = AirflowDagRunFacet(
        dag=expected_dag_info,
        dagRun={
            "conf": {},
            "dag_id": "dag_id",
            "data_interval_start": event_time.isoformat(),
            "data_interval_end": event_time.isoformat(),
            "external_trigger": False if AIRFLOW_V_3_0_PLUS else None,
            "run_id": run_id,
            "run_type": DagRunType.MANUAL,
            "start_date": event_time.isoformat(),
        },
    )
    adapter.dag_started(
        dag_id=dag_id,
        start_date=event_time,
        logical_date=event_time,
        clear_number=0,
        nominal_start_time=event_time.isoformat(),
        nominal_end_time=event_time.isoformat(),
        owners=["owner1", "owner2"],
        job_description=dag.description,
        job_description_type="text/plain",
        tags=["tag1", "tag2"],
        run_facets={
            "parent": parent_run.ParentRunFacet(
                run=parent_run.Run(runId=random_uuid),
                job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                root=parent_run.Root(
                    run=parent_run.RootRun(runId=random_uuid),
                    job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                ),
            ),
            "airflowDagRun": dag_run_facet,
        },
        job_facets=job_facets,
    )

    assert len(client.emit.mock_calls) == 1
    client.emit.assert_called_once_with(
        RunEvent(
            eventType=RunState.START,
            eventTime=event_time.isoformat(),
            run=Run(
                runId=random_uuid,
                facets={
                    "parent": parent_run.ParentRunFacet(
                        run=parent_run.Run(runId=random_uuid),
                        job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                        root=parent_run.Root(
                            run=parent_run.RootRun(runId=random_uuid),
                            job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                        ),
                    ),
                    "nominalTime": nominal_time_run.NominalTimeRunFacet(
                        nominalStartTime=event_time.isoformat(),
                        nominalEndTime=event_time.isoformat(),
                    ),
                    "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                        version=ANY, name="Airflow", openlineageAdapterVersion=ANY
                    ),
                    "airflowDagRun": AirflowDagRunFacet(
                        dag=expected_dag_info,
                        dagRun={
                            "conf": {},
                            "dag_id": "dag_id",
                            "data_interval_start": event_time.isoformat(),
                            "data_interval_end": event_time.isoformat(),
                            "external_trigger": False if AIRFLOW_V_3_0_PLUS else None,
                            "run_id": run_id,
                            "run_type": DagRunType.MANUAL,
                            "start_date": event_time.isoformat(),
                        },
                    ),
                    "debug": AirflowDebugRunFacet(packages=ANY),
                },
            ),
            job=Job(
                namespace=namespace(),
                name="dag_id",
                facets={
                    "documentation": documentation_job.DocumentationJobFacet(
                        description="dag desc", contentType="text/plain"
                    ),
                    "ownership": ownership_job.OwnershipJobFacet(
                        owners=[
                            ownership_job.Owner(name="owner1", type=None),
                            ownership_job.Owner(name="owner2", type=None),
                        ]
                    ),
                    "tags": tags_job.TagsJobFacet(
                        tags=[
                            tags_job.TagsJobFacetFields(key="tag1", value="tag1", source="AIRFLOW"),
                            tags_job.TagsJobFacetFields(key="tag2", value="tag2", source="AIRFLOW"),
                        ]
                    ),
                    **job_facets,
                    "jobType": job_type_job.JobTypeJobFacet(
                        processingType="BATCH", integration="AIRFLOW", jobType="DAG"
                    ),
                },
            ),
            producer=_PRODUCER,
            inputs=[],
            outputs=[],
        )
    )
    mock_stats_incr.assert_not_called()
    mock_stats_timer.assert_called_with("ol.emit.attempts")


@mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
@mock.patch.object(DagRun, "fetch_task_instances")
@mock.patch("airflow.providers.openlineage.plugins.adapter.generate_static_uuid")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.timer")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.incr")
def test_emit_dag_complete_event(
    mock_stats_incr, mock_stats_timer, generate_static_uuid, mocked_fetch_tis, mock_debug_mode
):
    random_uuid = "9d3b14f7-de91-40b6-aeef-e887e2c7673e"
    client = MagicMock()
    adapter = OpenLineageAdapter(client)
    event_time = datetime.datetime.now()
    dag_id = "dag_id"
    run_id = str(uuid.uuid4())

    with DAG(dag_id=dag_id, schedule=None, start_date=datetime.datetime(2024, 6, 1)):
        task_0 = BashOperator(task_id="task_0", bash_command="exit 0;")
        task_1 = BashOperator(task_id="task_1", bash_command="exit 0;")
        task_2 = EmptyOperator(
            task_id="task_2.test",
        )
    if AIRFLOW_V_3_0_PLUS:
        dag_run = DagRun(
            dag_id=dag_id,
            run_id=run_id,
            start_date=event_time,
            logical_date=event_time,
        )
    else:
        dag_run = DagRun(
            dag_id=dag_id,
            run_id=run_id,
            start_date=event_time,
            execution_date=event_time,
        )

    dag_run._state = DagRunState.SUCCESS
    dag_run.end_date = event_time
    if AIRFLOW_V_3_0_PLUS:
        ti0 = create_task_instance(
            task=task_0, run_id=run_id, state=TaskInstanceState.SUCCESS, dag_version_id=mock.MagicMock()
        )
        ti1 = create_task_instance(
            task=task_1, run_id=run_id, state=TaskInstanceState.SKIPPED, dag_version_id=mock.MagicMock()
        )

        ti2 = create_task_instance(
            task=task_2, run_id=run_id, state=TaskInstanceState.FAILED, dag_version_id=mock.MagicMock()
        )
    else:
        ti0 = TaskInstance(task=task_0, run_id=run_id, state=TaskInstanceState.SUCCESS)
        ti1 = TaskInstance(task=task_1, run_id=run_id, state=TaskInstanceState.SKIPPED)
        ti2 = TaskInstance(task=task_2, run_id=run_id, state=TaskInstanceState.FAILED)

    ti0.start_date = datetime.datetime(2022, 1, 1, 0, 0, 0)
    ti0.end_date = datetime.datetime(2022, 1, 1, 0, 10, 0)

    ti1.start_date = datetime.datetime(2022, 1, 1, 0, 10, 2)
    ti1.end_date = datetime.datetime(2022, 1, 1, 0, 13, 7)

    ti2.start_date = datetime.datetime(2022, 1, 1, 0, 13, 8)
    ti2.end_date = datetime.datetime(2022, 1, 1, 0, 14, 0)

    mocked_fetch_tis.return_value = [ti0, ti1, ti2]
    generate_static_uuid.return_value = random_uuid

    adapter.dag_success(
        dag_id=dag_id,
        run_id=run_id,
        end_date=event_time,
        logical_date=event_time,
        clear_number=0,
        dag_run_state=DagRunState.SUCCESS,
        task_ids=["task_0", "task_1", "task_2.test"],
        owners=["owner1", "owner2"],
        tags=["tag1", "tag2"],
        job_description="dag desc",
        job_description_type="text/plain",
        nominal_start_time=datetime.datetime(2022, 1, 1).isoformat(),
        nominal_end_time=datetime.datetime(2022, 1, 1).isoformat(),
        run_facets={
            "parent": parent_run.ParentRunFacet(
                run=parent_run.Run(runId=random_uuid),
                job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                root=parent_run.Root(
                    run=parent_run.RootRun(runId=random_uuid),
                    job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                ),
            ),
            "airflowDagRun": AirflowDagRunFacet(dag={"description": "dag desc"}, dagRun=dag_run),
        },
    )

    client.emit.assert_called_once_with(
        RunEvent(
            eventType=RunState.COMPLETE,
            eventTime=event_time.isoformat(),
            run=Run(
                runId=random_uuid,
                facets={
                    "parent": parent_run.ParentRunFacet(
                        run=parent_run.Run(runId=random_uuid),
                        job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                        root=parent_run.Root(
                            run=parent_run.RootRun(runId=random_uuid),
                            job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                        ),
                    ),
                    "nominalTime": nominal_time_run.NominalTimeRunFacet(
                        nominalStartTime="2022-01-01T00:00:00",
                        nominalEndTime="2022-01-01T00:00:00",
                    ),
                    "airflowState": AirflowStateRunFacet(
                        dagRunState=DagRunState.SUCCESS,
                        tasksState={
                            task_0.task_id: TaskInstanceState.SUCCESS,
                            task_1.task_id: TaskInstanceState.SKIPPED,
                            task_2.task_id: TaskInstanceState.FAILED,
                        },
                        tasksDuration={
                            task_0.task_id: 600.0,
                            task_1.task_id: 185.0,
                            task_2.task_id: 52.0,
                        },
                    ),
                    "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                        version=ANY, name="Airflow", openlineageAdapterVersion=ANY
                    ),
                    "debug": AirflowDebugRunFacet(packages=ANY),
                    "airflowDagRun": AirflowDagRunFacet(dag={"description": "dag desc"}, dagRun=dag_run),
                },
            ),
            job=Job(
                namespace=namespace(),
                name=dag_id,
                facets={
                    "documentation": documentation_job.DocumentationJobFacet(
                        description="dag desc", contentType="text/plain"
                    ),
                    "ownership": ownership_job.OwnershipJobFacet(
                        owners=[
                            ownership_job.Owner(name="owner1", type=None),
                            ownership_job.Owner(name="owner2", type=None),
                        ]
                    ),
                    "tags": tags_job.TagsJobFacet(
                        tags=[
                            tags_job.TagsJobFacetFields(key="tag1", value="tag1", source="AIRFLOW"),
                            tags_job.TagsJobFacetFields(key="tag2", value="tag2", source="AIRFLOW"),
                        ]
                    ),
                    "jobType": job_type_job.JobTypeJobFacet(
                        processingType="BATCH", integration="AIRFLOW", jobType="DAG"
                    ),
                },
            ),
            producer=_PRODUCER,
            inputs=[],
            outputs=[],
        )
    )

    mock_stats_incr.assert_not_called()
    mock_stats_timer.assert_called_with("ol.emit.attempts")


@mock.patch("airflow.providers.openlineage.conf.debug_mode", return_value=True)
@mock.patch.object(DagRun, "fetch_task_instances")
@mock.patch("airflow.providers.openlineage.plugins.adapter.generate_static_uuid")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.timer")
@mock.patch("airflow.providers.openlineage.plugins.adapter.Stats.incr")
def test_emit_dag_failed_event(
    mock_stats_incr, mock_stats_timer, generate_static_uuid, mocked_fetch_tis, mock_debug_mode
):
    random_uuid = "9d3b14f7-de91-40b6-aeef-e887e2c7673e"
    client = MagicMock()
    adapter = OpenLineageAdapter(client)
    event_time = datetime.datetime.now()
    dag_id = "dag_id"
    run_id = str(uuid.uuid4())

    with DAG(dag_id=dag_id, schedule=None, start_date=datetime.datetime(2024, 6, 1)):
        task_0 = BashOperator(task_id="task_0", bash_command="exit 0;")
        task_1 = BashOperator(task_id="task_1", bash_command="exit 0;")
        task_2 = EmptyOperator(task_id="task_2.test")

    if AIRFLOW_V_3_0_PLUS:
        dag_run = DagRun(
            dag_id=dag_id,
            run_id=run_id,
            start_date=event_time,
            logical_date=event_time,
        )
    else:
        dag_run = DagRun(
            dag_id=dag_id,
            run_id=run_id,
            start_date=event_time,
            execution_date=event_time,
        )
    dag_run._state = DagRunState.FAILED
    dag_run.end_date = event_time

    if AIRFLOW_V_3_0_PLUS:
        ti0 = create_task_instance(
            task=task_0, run_id=run_id, state=TaskInstanceState.SUCCESS, dag_version_id=mock.MagicMock()
        )
        ti1 = create_task_instance(
            task=task_1, run_id=run_id, state=TaskInstanceState.SKIPPED, dag_version_id=mock.MagicMock()
        )

        ti2 = create_task_instance(
            task=task_2, run_id=run_id, state=TaskInstanceState.FAILED, dag_version_id=mock.MagicMock()
        )
    else:
        ti0 = TaskInstance(task=task_0, run_id=run_id, state=TaskInstanceState.SUCCESS)
        ti1 = TaskInstance(task=task_1, run_id=run_id, state=TaskInstanceState.SKIPPED)
        ti2 = TaskInstance(task=task_2, run_id=run_id, state=TaskInstanceState.FAILED)

    ti0.start_date = datetime.datetime(2022, 1, 1, 0, 0, 0)
    ti0.end_date = datetime.datetime(2022, 1, 1, 0, 10, 0)

    ti1.start_date = datetime.datetime(2022, 1, 1, 0, 10, 2)
    ti1.end_date = datetime.datetime(2022, 1, 1, 0, 13, 7)

    ti2.start_date = datetime.datetime(2022, 1, 1, 0, 13, 8)
    ti2.end_date = datetime.datetime(2022, 1, 1, 0, 14, 0)

    mocked_fetch_tis.return_value = [ti0, ti1, ti2]

    generate_static_uuid.return_value = random_uuid

    adapter.dag_failed(
        dag_id=dag_id,
        run_id=run_id,
        end_date=event_time,
        logical_date=event_time,
        clear_number=0,
        dag_run_state=DagRunState.FAILED,
        task_ids=["task_0", "task_1", "task_2.test"],
        tags=["tag1", "tag2"],
        msg="error msg",
        owners=["owner1", "owner2"],
        job_description="dag desc",
        job_description_type="text/plain",
        nominal_start_time=datetime.datetime(2022, 1, 1).isoformat(),
        nominal_end_time=datetime.datetime(2022, 1, 1).isoformat(),
        run_facets={
            "parent": parent_run.ParentRunFacet(
                run=parent_run.Run(runId=random_uuid),
                job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                root=parent_run.Root(
                    run=parent_run.RootRun(runId=random_uuid),
                    job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                ),
            ),
            "airflowDagRun": AirflowDagRunFacet(dag={"description": "dag desc"}, dagRun=dag_run),
        },
    )

    client.emit.assert_called_once_with(
        RunEvent(
            eventType=RunState.FAIL,
            eventTime=event_time.isoformat(),
            run=Run(
                runId=random_uuid,
                facets={
                    "parent": parent_run.ParentRunFacet(
                        run=parent_run.Run(runId=random_uuid),
                        job=parent_run.Job(namespace=namespace(), name="parent_job_name"),
                        root=parent_run.Root(
                            run=parent_run.RootRun(runId=random_uuid),
                            job=parent_run.RootJob(namespace=namespace(), name="parent_job_name"),
                        ),
                    ),
                    "nominalTime": nominal_time_run.NominalTimeRunFacet(
                        nominalStartTime="2022-01-01T00:00:00",
                        nominalEndTime="2022-01-01T00:00:00",
                    ),
                    "errorMessage": error_message_run.ErrorMessageRunFacet(
                        message="error msg", programmingLanguage="python"
                    ),
                    "airflowState": AirflowStateRunFacet(
                        dagRunState=DagRunState.FAILED,
                        tasksState={
                            task_0.task_id: TaskInstanceState.SUCCESS,
                            task_1.task_id: TaskInstanceState.SKIPPED,
                            task_2.task_id: TaskInstanceState.FAILED,
                        },
                        tasksDuration={
                            task_0.task_id: 600.0,
                            task_1.task_id: 185.0,
                            task_2.task_id: 52.0,
                        },
                    ),
                    "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
                        version=ANY, name="Airflow", openlineageAdapterVersion=ANY
                    ),
                    "debug": AirflowDebugRunFacet(packages=ANY),
                    "airflowDagRun": AirflowDagRunFacet(dag={"description": "dag desc"}, dagRun=dag_run),
                },
            ),
            job=Job(
                namespace=namespace(),
                name=dag_id,
                facets={
                    "documentation": documentation_job.DocumentationJobFacet(
                        description="dag desc", contentType="text/plain"
                    ),
                    "ownership": ownership_job.OwnershipJobFacet(
                        owners=[
                            ownership_job.Owner(name="owner1", type=None),
                            ownership_job.Owner(name="owner2", type=None),
                        ]
                    ),
                    "tags": tags_job.TagsJobFacet(
                        tags=[
                            tags_job.TagsJobFacetFields(key="tag1", value="tag1", source="AIRFLOW"),
                            tags_job.TagsJobFacetFields(key="tag2", value="tag2", source="AIRFLOW"),
                        ]
                    ),
                    "jobType": job_type_job.JobTypeJobFacet(
                        processingType="BATCH", integration="AIRFLOW", jobType="DAG"
                    ),
                },
            ),
            producer=_PRODUCER,
            inputs=[],
            outputs=[],
        )
    )

    mock_stats_incr.assert_not_called()
    mock_stats_timer.assert_called_with("ol.emit.attempts")


@patch("airflow.providers.openlineage.plugins.adapter.OpenLineageAdapter.get_or_create_openlineage_client")
@patch("airflow.providers.openlineage.plugins.adapter.OpenLineageRedactor")
@patch("airflow.providers.openlineage.plugins.adapter.Stats.timer")
@patch("airflow.providers.openlineage.plugins.adapter.Stats.incr")
def test_openlineage_adapter_stats_emit_failed(
    mock_stats_incr, mock_stats_timer, mock_redact, mock_get_client
):
    adapter = OpenLineageAdapter()
    mock_get_client.return_value.emit.side_effect = Exception()

    adapter.emit(MagicMock())

    mock_stats_timer.assert_called_with("ol.emit.attempts")
    mock_stats_incr.assert_has_calls([mock.call("ol.emit.failed")])


def test_build_dag_run_id_is_valid_uuid():
    dag_id = "test_dag"
    logical_date = datetime.datetime.now()
    result = OpenLineageAdapter.build_dag_run_id(
        dag_id=dag_id,
        logical_date=logical_date,
        clear_number=0,
    )
    uuid_result = uuid.UUID(result)
    assert uuid_result
    assert uuid_result.version == 7


def test_build_dag_run_id_same_input_give_same_result():
    result1 = OpenLineageAdapter.build_dag_run_id(
        dag_id="dag1", logical_date=datetime.datetime(2024, 1, 1, 1, 1, 1), clear_number=0
    )
    result2 = OpenLineageAdapter.build_dag_run_id(
        dag_id="dag1", logical_date=datetime.datetime(2024, 1, 1, 1, 1, 1), clear_number=0
    )
    assert result1 == result2


def test_build_dag_run_id_different_inputs_give_different_results():
    result1 = OpenLineageAdapter.build_dag_run_id(
        dag_id="dag1", logical_date=datetime.datetime.now(), clear_number=0
    )
    result2 = OpenLineageAdapter.build_dag_run_id(
        dag_id="dag2", logical_date=datetime.datetime.now(), clear_number=0
    )
    assert result1 != result2


def test_build_dag_run_id_different_clear_number_give_different_results():
    result1 = OpenLineageAdapter.build_dag_run_id(
        dag_id="dag1", logical_date=datetime.datetime(2024, 1, 1, 1, 1, 1), clear_number=0
    )
    result2 = OpenLineageAdapter.build_dag_run_id(
        dag_id="dag1", logical_date=datetime.datetime(2024, 1, 1, 1, 1, 1), clear_number=1
    )
    assert result1 != result2


def test_build_task_instance_run_id_is_valid_uuid():
    result = OpenLineageAdapter.build_task_instance_run_id(
        dag_id="dag_id",
        task_id="task_id",
        try_number=1,
        logical_date=datetime.datetime.now(),
        map_index=-1,
    )
    uuid_result = uuid.UUID(result)
    assert uuid_result
    assert uuid_result.version == 7


def test_build_task_instance_run_id_same_input_gives_same_result():
    result1 = OpenLineageAdapter.build_task_instance_run_id(
        dag_id="dag1",
        task_id="task1",
        try_number=1,
        logical_date=datetime.datetime(2024, 1, 1, 1, 1, 1),
        map_index=-1,
    )
    result2 = OpenLineageAdapter.build_task_instance_run_id(
        dag_id="dag1",
        task_id="task1",
        try_number=1,
        logical_date=datetime.datetime(2024, 1, 1, 1, 1, 1),
        map_index=-1,
    )
    assert result1 == result2


def test_build_task_instance_run_id_different_map_index_gives_different_result():
    result1 = OpenLineageAdapter.build_task_instance_run_id(
        dag_id="dag1",
        task_id="task1",
        try_number=1,
        logical_date=datetime.datetime(2024, 1, 1, 1, 1, 1),
        map_index=1,
    )
    result2 = OpenLineageAdapter.build_task_instance_run_id(
        dag_id="dag1",
        task_id="task1",
        try_number=1,
        logical_date=datetime.datetime(2024, 1, 1, 1, 1, 1),
        map_index=2,
    )
    assert result1 != result2


def test_build_task_instance_run_id_different_inputs_gives_different_results():
    result1 = OpenLineageAdapter.build_task_instance_run_id(
        dag_id="dag1",
        task_id="task1",
        try_number=1,
        logical_date=datetime.datetime.now(),
        map_index=-1,
    )
    result2 = OpenLineageAdapter.build_task_instance_run_id(
        dag_id="dag2",
        task_id="task2",
        try_number=2,
        logical_date=datetime.datetime.now(),
        map_index=-1,
    )
    assert result1 != result2


@skip_if_force_lowest_dependencies_marker
def test_configuration_precedence_when_creating_ol_client():
    _section_name = "openlineage"
    current_folder = pathlib.Path(__file__).parent.resolve()
    yaml_config = str((current_folder / "openlineage_configs" / "http.yaml").resolve())

    # First, check config_path in Airflow configuration (airflow.cfg or env variable)
    with patch.dict(
        os.environ,
        {
            "OPENLINEAGE_URL": "http://wrong.com",
            "OPENLINEAGE_API_KEY": "wrong_api_key",
            "OPENLINEAGE_CONFIG": "some/config.yml",
        },
        clear=True,
    ):
        with conf_vars(
            {
                (_section_name, "transport"): '{"type": "kafka", "topic": "test", "config": {"acks": "all"}}',
                (_section_name, "config_path"): yaml_config,
            },
        ):
            client = OpenLineageAdapter().get_or_create_openlineage_client()
            assert client.transport.kind == "http"
            assert client.transport.config.url == "http://localhost:5050"
            assert client.transport.config.endpoint == "api/v1/lineage"
            assert client.transport.config.auth.api_key == "random_token"

    # Second, check transport in Airflow configuration (airflow.cfg or env variable)
    with patch.dict(
        os.environ,
        {
            "OPENLINEAGE_URL": "http://wrong.com",
            "OPENLINEAGE_API_KEY": "wrong_api_key",
            "OPENLINEAGE_CONFIG": "some/config.yml",
        },
        clear=True,
    ):
        with conf_vars(
            {
                (_section_name, "transport"): '{"type": "kafka", "topic": "test", "config": {"acks": "all"}}',
                (_section_name, "config_path"): "",
            },
        ):
            client = OpenLineageAdapter().get_or_create_openlineage_client()
            assert client.transport.kind == "kafka"
            assert client.transport.kafka_config.topic == "test"
            assert client.transport.kafka_config.config == {"acks": "all"}

    # Third, check legacy OPENLINEAGE_CONFIG env variable
    with patch.dict(
        os.environ,
        {
            "OPENLINEAGE_URL": "http://wrong.com",
            "OPENLINEAGE_API_KEY": "wrong_api_key",
            "OPENLINEAGE_CONFIG": yaml_config,
        },
        clear=True,
    ):
        with conf_vars(
            {
                (_section_name, "transport"): "",
                (_section_name, "config_path"): "",
            },
        ):
            client = OpenLineageAdapter().get_or_create_openlineage_client()
            assert client.transport.kind == "http"
            assert client.transport.config.url == "http://localhost:5050"
            assert client.transport.config.endpoint == "api/v1/lineage"
            assert client.transport.config.auth.api_key == "random_token"

    # Fourth, check legacy OPENLINEAGE_URL env variable
    with patch.dict(
        os.environ,
        {
            "OPENLINEAGE_URL": "http://test.com",
            "OPENLINEAGE_API_KEY": "test_api_key",
            "OPENLINEAGE_CONFIG": "",
        },
        clear=True,
    ):
        with conf_vars(
            {
                (_section_name, "transport"): "",
                (_section_name, "config_path"): "",
            },
        ):
            client = OpenLineageAdapter().get_or_create_openlineage_client()
            assert client.transport.kind == "http"
            assert client.transport.config.url == "http://test.com"
            assert client.transport.config.endpoint == "api/v1/lineage"
            assert client.transport.config.auth.api_key == "test_api_key"

    # If all else fails, use console transport
    with patch.dict(os.environ, {}, clear=True):
        with conf_vars(
            {
                (_section_name, "transport"): "",
                (_section_name, "config_path"): "",
            },
        ):
            client = OpenLineageAdapter().get_or_create_openlineage_client()
            assert client.transport.kind == "console"


def test_adapter_build_run():
    run_id = str(uuid.uuid4())
    result = OpenLineageAdapter._build_run(
        run_id=run_id,
        nominal_start_time=datetime.datetime(2022, 1, 1).isoformat(),
        nominal_end_time=datetime.datetime(2022, 1, 1).isoformat(),
        run_facets={
            "my_custom_facet": external_query_run.ExternalQueryRunFacet(
                externalQueryId="123", source="source"
            ),
            "processing_engine": "this_should_be_gone",
        },
    )
    assert result.runId == run_id
    assert result.facets == {
        "my_custom_facet": external_query_run.ExternalQueryRunFacet(externalQueryId="123", source="source"),
        "nominalTime": nominal_time_run.NominalTimeRunFacet(
            nominalStartTime="2022-01-01T00:00:00",
            nominalEndTime="2022-01-01T00:00:00",
        ),
        "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
            version=ANY, name="Airflow", openlineageAdapterVersion=ANY
        ),
    }


def test_adapter_build_job():
    result = OpenLineageAdapter._build_job(
        job_name="job_name",
        job_type="TASK",
        job_description="job_description",
        job_owners=["def", "abc"],
        job_tags=["tag2", "tag1"],
        job_facets={
            "my_custom_facet": sql_job.SQLJobFacet(query="sql"),
            "jobType": "this_should_be_gone",
            "documentation": "this_should_be_gone",
            "ownership": "this_should_be_gone",
            "tags": "this_should_be_gone",
        },
    )
    assert result.name == "job_name"
    assert result.facets == {
        "my_custom_facet": sql_job.SQLJobFacet(query="sql"),
        "documentation": documentation_job.DocumentationJobFacet(description="job_description"),
        "ownership": ownership_job.OwnershipJobFacet(
            owners=[
                ownership_job.Owner(name="abc", type=None),
                ownership_job.Owner(name="def", type=None),
            ]
        ),
        "tags": tags_job.TagsJobFacet(
            tags=[
                tags_job.TagsJobFacetFields(key="tag1", value="tag1", source="AIRFLOW"),
                tags_job.TagsJobFacetFields(key="tag2", value="tag2", source="AIRFLOW"),
            ]
        ),
        "jobType": job_type_job.JobTypeJobFacet(
            processingType="BATCH", integration="AIRFLOW", jobType="TASK"
        ),
    }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Simple DAG with mapped task.

It checks:
    - task's `mapped` attribute
    - taskInstance's `map_index` attribute
    - number of OL events emitted for mapped task
"""

from __future__ import annotations

from datetime import datetime

from airflow import DAG

try:
    from airflow.sdk import task
except ImportError:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models import Variable
from airflow.providers.standard.operators.python import PythonOperator

from system.openlineage.expected_events import get_expected_event_file_path
from system.openlineage.operator import OpenLineageTestOperator


def check_events_number_func():
    for event_type in ("start", "complete"):
        events = Variable.get(
            key=f"openlineage_mapped_simple_dag.add_one.event.{event_type}", deserialize_json=True
        )
        if len(events) != 2:
            raise ValueError(
                f"Expected exactly 2 {event_type.upper()} events for task `add_one`, got {len(events)}"
            )


DAG_ID = "openlineage_mapped_simple_dag"

with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule=None,
    catchup=False,
    default_args={"retries": 0},
) as dag:

    @task(max_active_tis_per_dagrun=1)  # Execute sequentially to not overwrite each other OL events
    def add_one(x: int):
        return x + 1

    @task
    def sum_it(values):
        total = sum(values)
        print(f"Total was {total}")

    added_values = add_one.expand(x=[1, 2])

    check_events_number = PythonOperator(
        task_id="check_events_number", python_callable=check_events_number_func
    )

    check_events = OpenLineageTestOperator(
        task_id="check_events",
        file_path=get_expected_event_file_path(DAG_ID),
        allow_duplicate_events_regex="openlineage_mapped_simple_dag.add_one.event.(start|complete)",
    )

    sum_it(added_values) >> check_events_number >> check_events


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import datetime
import json
import logging
from collections.abc import Callable
from contextlib import suppress
from functools import wraps
from importlib import metadata
from typing import TYPE_CHECKING, Any

import attrs
from openlineage.client.facet_v2 import parent_run
from openlineage.client.utils import RedactMixin

from airflow import __version__ as AIRFLOW_VERSION

# TODO: move this maybe to Airflow's logic?
from airflow.models import DagRun, TaskInstance, TaskReschedule
from airflow.providers.common.compat.assets import Asset
from airflow.providers.common.compat.module_loading import import_string
from airflow.providers.common.compat.sdk import DAG, BaseOperator, BaseSensorOperator, MappedOperator
from airflow.providers.openlineage import (
    __version__ as OPENLINEAGE_PROVIDER_VERSION,
    conf,
)
from airflow.providers.openlineage.plugins.facets import (
    AirflowDagRunFacet,
    AirflowDebugRunFacet,
    AirflowJobFacet,
    AirflowMappedTaskRunFacet,
    AirflowRunFacet,
    AirflowStateRunFacet,
    UnknownOperatorAttributeRunFacet,
    UnknownOperatorInstance,
)
from airflow.providers.openlineage.utils.selective_enable import (
    is_dag_lineage_enabled,
    is_task_lineage_enabled,
)
from airflow.providers.openlineage.version_compat import AIRFLOW_V_3_0_PLUS, get_base_airflow_version_tuple
from airflow.serialization.serialized_objects import SerializedBaseOperator, SerializedDAG

try:
    from airflow.serialization.definitions.mappedoperator import SerializedMappedOperator
except ImportError:
    from airflow.models.mappedoperator import (  # type: ignore[attr-defined,no-redef]
        MappedOperator as SerializedMappedOperator,
    )

if not AIRFLOW_V_3_0_PLUS:
    from airflow.utils.session import NEW_SESSION, provide_session

if TYPE_CHECKING:
    from typing import TypeAlias

    from openlineage.client.event_v2 import Dataset as OpenLineageDataset
    from openlineage.client.facet_v2 import RunFacet, processing_engine_run

    from airflow.sdk.execution_time.secrets_masker import (
        Redactable,
        Redacted,
        SecretsMasker,
    )
    from airflow.sdk.execution_time.task_runner import RuntimeTaskInstance
    from airflow.utils.state import DagRunState, TaskInstanceState

    AnyOperator: TypeAlias = BaseOperator | MappedOperator | SerializedBaseOperator | SerializedMappedOperator
else:
    try:
        from airflow.sdk._shared.secrets_masker import (
            Redactable,
            Redacted,
            SecretsMasker,
            should_hide_value_for_key,
        )
    except ImportError:
        try:
            from airflow.sdk.execution_time.secrets_masker import (
                Redactable,
                Redacted,
                SecretsMasker,
                should_hide_value_for_key,
            )
        except ImportError:
            from airflow.utils.log.secrets_masker import (
                Redactable,
                Redacted,
                SecretsMasker,
                should_hide_value_for_key,
            )

log = logging.getLogger(__name__)
_NOMINAL_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
_MAX_DOC_BYTES = 64 * 1024  # 64 kilobytes


def try_import_from_string(string: str) -> Any:
    with suppress(ImportError):
        return import_string(string)


def get_operator_class(task: BaseOperator | MappedOperator) -> type:
    if task.__class__.__name__ in ("DecoratedMappedOperator", "MappedOperator"):
        return task.operator_class
    return task.__class__


def get_operator_provider_version(operator: AnyOperator) -> str | None:
    """Get the provider package version for the given operator."""
    try:
        class_path = get_fully_qualified_class_name(operator)

        if not class_path.startswith("airflow.providers."):
            return None

        from airflow.providers_manager import ProvidersManager

        providers_manager = ProvidersManager()

        for package_name, provider_info in providers_manager.providers.items():
            if package_name.startswith("apache-airflow-providers-"):
                provider_module_path = package_name.replace(
                    "apache-airflow-providers-", "airflow.providers."
                ).replace("-", ".")
                if class_path.startswith(provider_module_path + "."):
                    return provider_info.version

        return None

    except Exception:
        return None


def get_job_name(task: TaskInstance | RuntimeTaskInstance) -> str:
    return f"{task.dag_id}.{task.task_id}"


def _get_parent_run_facet(
    parent_run_id: str,
    parent_job_name: str,
    parent_job_namespace: str = conf.namespace(),
    root_parent_run_id: str | None = None,
    root_parent_job_name: str | None = None,
    root_parent_job_namespace: str | None = None,
) -> dict[str, Any]:
    """Create the parent run facet from identifiers."""
    root_parent_run_id = root_parent_run_id or parent_run_id
    root_parent_job_name = root_parent_job_name or parent_job_name
    root_parent_job_namespace = root_parent_job_namespace or parent_job_namespace
    return {
        "parent": parent_run.ParentRunFacet(
            run=parent_run.Run(runId=parent_run_id),
            job=parent_run.Job(namespace=parent_job_namespace, name=parent_job_name),
            root=parent_run.Root(
                run=parent_run.RootRun(runId=root_parent_run_id),
                job=parent_run.RootJob(namespace=root_parent_job_namespace, name=root_parent_job_name),
            ),
        )
    }


def get_task_parent_run_facet(
    parent_run_id: str,
    parent_job_name: str,
    parent_job_namespace: str = conf.namespace(),
    root_parent_run_id: str | None = None,
    root_parent_job_name: str | None = None,
    root_parent_job_namespace: str | None = None,
    dr_conf: dict | None = None,
) -> dict[str, Any]:
    """Retrieve the parent run facet."""
    all_root_info = (root_parent_run_id, root_parent_job_namespace, root_parent_job_name)
    # If not all root identifiers are provided explicitly, try to get them from dagrun conf.
    if not all(all_root_info):
        # If some but not all root identifiers are provided warn and do not use any of them, we need all.
        if any(all_root_info):
            log.warning(
                "Incomplete root OpenLineage information provided. "
                "No root information will be used. Found values: "
                "root_parent_run_id='%s', root_parent_job_namespace='%s', root_parent_job_name='%s'.",
                root_parent_run_id,
                root_parent_job_namespace,
                root_parent_job_name,
            )
            root_parent_run_id, root_parent_job_namespace, root_parent_job_name = None, None, None
        elif dr_conf:
            # Check for root identifiers in dagrun conf
            if root_info := get_root_information_from_dagrun_conf(dr_conf):
                root_parent_run_id = root_info["root_parent_run_id"]
                root_parent_job_namespace = root_info["root_parent_job_namespace"]
                root_parent_job_name = root_info["root_parent_job_name"]
            # If not present, check for parent identifiers in dagrun conf and use them as root
            elif parent_info := get_parent_information_from_dagrun_conf(dr_conf):
                root_parent_run_id = parent_info["parent_run_id"]
                root_parent_job_namespace = parent_info["parent_job_namespace"]
                root_parent_job_name = parent_info["parent_job_name"]

    return _get_parent_run_facet(
        parent_run_id=parent_run_id,
        parent_job_namespace=parent_job_namespace,
        parent_job_name=parent_job_name,
        root_parent_run_id=root_parent_run_id,
        root_parent_job_namespace=root_parent_job_namespace,
        root_parent_job_name=root_parent_job_name,
    )


def _get_openlineage_data_from_dagrun_conf(dr_conf: dict | None) -> dict:
    """Return the 'openlineage' section from a DAG run config if valid, otherwise an empty dict."""
    if not dr_conf or not isinstance(dr_conf, dict):
        return {}
    ol_data = dr_conf.get("openlineage")
    return ol_data if isinstance(ol_data, dict) else {}


def get_root_information_from_dagrun_conf(dr_conf: dict | None) -> dict[str, str]:
    """Extract root parent run and job information from a DAG run config."""
    ol_data = _get_openlineage_data_from_dagrun_conf(dr_conf)
    if not ol_data:
        log.debug("No 'openlineage' data found in DAG run config.")
        return {}

    root_parent_run_id = ol_data.get("rootParentRunId", "")
    root_parent_job_namespace = ol_data.get("rootParentJobNamespace", "")
    root_parent_job_name = ol_data.get("rootParentJobName", "")

    all_root_info = (root_parent_run_id, root_parent_job_namespace, root_parent_job_name)
    if not all(all_root_info):
        if any(all_root_info):
            log.warning(
                "Incomplete root OpenLineage information in DAG run config. "
                "No root information will be used. Found values: "
                "rootParentRunId='%s', rootParentJobNamespace='%s', rootParentJobName='%s'.",
                root_parent_run_id,
                root_parent_job_namespace,
                root_parent_job_name,
            )
        else:
            log.debug("No 'openlineage' root information found in DAG run config.")
        return {}

    try:  # Validate that runId is correct UUID
        parent_run.RootRun(runId=root_parent_run_id)
    except ValueError:
        log.warning(
            "Invalid OpenLineage rootParentRunId '%s' in DAG run config - expected a valid UUID.",
            root_parent_run_id,
        )
        return {}

    log.debug(
        "Extracted valid root OpenLineage identifiers from DAG run config: "
        "rootParentRunId='%s', rootParentJobNamespace='%s', rootParentJobName='%s'.",
        root_parent_run_id,
        root_parent_job_namespace,
        root_parent_job_name,
    )
    return {
        "root_parent_run_id": root_parent_run_id,
        "root_parent_job_namespace": root_parent_job_namespace,
        "root_parent_job_name": root_parent_job_name,
    }


def get_parent_information_from_dagrun_conf(dr_conf: dict | None) -> dict[str, str]:
    """Extract parent run and job information from a DAG run config."""
    ol_data = _get_openlineage_data_from_dagrun_conf(dr_conf)
    if not ol_data:
        log.debug("No 'openlineage' data found in DAG run config.")
        return {}

    parent_run_id = ol_data.get("parentRunId", "")
    parent_job_namespace = ol_data.get("parentJobNamespace", "")
    parent_job_name = ol_data.get("parentJobName", "")

    all_parent_info = (parent_run_id, parent_job_namespace, parent_job_name)
    if not all(all_parent_info):
        if any(all_parent_info):
            log.warning(
                "Incomplete parent OpenLineage information in DAG run config. "
                "ParentRunFacet will NOT be created. Found values: "
                "parentRunId='%s', parentJobNamespace='%s', parentJobName='%s'.",
                parent_run_id,
                parent_job_namespace,
                parent_job_name,
            )
        else:
            log.debug("No 'openlineage' parent information found in DAG run config.")
        return {}

    try:  # Validate that runId is correct UUID
        parent_run.RootRun(runId=parent_run_id)
    except ValueError:
        log.warning(
            "Invalid OpenLineage parentRunId '%s' in DAG run config - expected a valid UUID.",
            parent_run_id,
        )
        return {}

    log.debug(
        "Extracted valid parent OpenLineage identifiers from DAG run config: "
        "parentRunId='%s', parentJobNamespace='%s', parentJobName='%s'.",
        parent_run_id,
        parent_job_namespace,
        parent_job_name,
    )

    return {
        "parent_run_id": parent_run_id,
        "parent_job_namespace": parent_job_namespace,
        "parent_job_name": parent_job_name,
    }


def get_dag_parent_run_facet(dr_conf: dict | None) -> dict[str, parent_run.ParentRunFacet]:
    """
    Build the OpenLineage parent run facet from a DAG run configuration.

    This function extracts parent run identifiers - run ID, job namespace, and job name -
    from the DAG run configuration to construct an OpenLineage `ParentRunFacet`. It requires
    a complete set of parent identifiers to proceed; if some but not all are present, or if
    the run ID is invalid, the function returns an empty dictionary.

    When valid parent identifiers are found, it also attempts to retrieve corresponding
    root identifiers using `get_root_information_from_dagrun_conf`, which may fall back
    to the parent identifiers if no explicit root data is available. The resulting facet
    links both the immediate parent and root lineage information for the run.

    Args:
        dr_conf: The DAG run configuration dictionary.

    Returns:
        A dictionary containing a single entry mapping the facet name to the constructed
        `ParentRunFacet`. Returns an empty dictionary if the configuration does not contain
        a complete or valid set of parent identifiers.
    """
    ol_data = _get_openlineage_data_from_dagrun_conf(dr_conf)
    if not ol_data:
        log.debug("No 'openlineage' data found in DAG run config.")
        return {}

    parent_info = get_parent_information_from_dagrun_conf(dr_conf)
    if not parent_info:  # Validation and logging is done in the above function
        return {}

    root_parent_run_id, root_parent_job_namespace, root_parent_job_name = None, None, None
    if root_info := get_root_information_from_dagrun_conf(dr_conf):
        root_parent_run_id = root_info["root_parent_run_id"]
        root_parent_job_namespace = root_info["root_parent_job_namespace"]
        root_parent_job_name = root_info["root_parent_job_name"]
    else:
        log.debug(
            "Missing OpenLineage root identifiers in DAG run config, "
            "parent identifiers will be used as root instead."
        )

    # Function already uses parent as root if root is missing, no need to explicitly pass it
    return _get_parent_run_facet(
        parent_run_id=parent_info["parent_run_id"],
        parent_job_namespace=parent_info["parent_job_namespace"],
        parent_job_name=parent_info["parent_job_name"],
        root_parent_run_id=root_parent_run_id,
        root_parent_job_namespace=root_parent_job_namespace,
        root_parent_job_name=root_parent_job_name,
    )


def _truncate_string_to_byte_size(s: str, max_size: int = _MAX_DOC_BYTES) -> str:
    """
    Truncate a string to a maximum UTF-8 byte size, ensuring valid encoding.

    This is used to safely limit the size of string content (e.g., for OpenLineage events)
    without breaking multibyte characters. If truncation occurs, the result is a valid
    UTF-8 string with any partial characters at the end removed.

    Args:
        s (str): The input string to truncate.
        max_size (int): Maximum allowed size in bytes after UTF-8 encoding.

    Returns:
        str: A UTF-8-safe truncated string within the specified byte limit.
    """
    encoded = s.encode("utf-8")
    if len(encoded) <= max_size:
        return s
    log.debug(
        "Truncating long string content for OpenLineage event. "
        "Original size: %d bytes, truncated to: %d bytes (UTF-8 safe).",
        len(encoded),
        max_size,
    )
    truncated = encoded[:max_size]
    # Make sure we don't cut a multibyte character in half
    return truncated.decode("utf-8", errors="ignore")


def get_task_documentation(operator: AnyOperator | None) -> tuple[str | None, str | None]:
    """Get task documentation and mime type, truncated to _MAX_DOC_BYTES bytes length, if present."""
    if not operator:
        return None, None

    doc, mime_type = None, None
    if operator.doc:
        doc = operator.doc
        mime_type = "text/plain"
    elif operator.doc_md:
        doc = operator.doc_md
        mime_type = "text/markdown"
    elif operator.doc_json:
        doc = operator.doc_json
        mime_type = "application/json"
    elif operator.doc_yaml:
        doc = operator.doc_yaml
        mime_type = "application/x-yaml"
    elif operator.doc_rst:
        doc = operator.doc_rst
        mime_type = "text/x-rst"

    if doc:
        return _truncate_string_to_byte_size(doc), mime_type
    return None, None


def get_dag_documentation(dag: DAG | SerializedDAG | None) -> tuple[str | None, str | None]:
    """Get dag documentation and mime type, truncated to _MAX_DOC_BYTES bytes length, if present."""
    if not dag:
        return None, None

    doc, mime_type = None, None
    if dag.doc_md:
        doc = dag.doc_md
        mime_type = "text/markdown"
    elif dag.description:
        doc = dag.description
        mime_type = "text/plain"

    if doc:
        return _truncate_string_to_byte_size(doc), mime_type
    return None, None


def get_airflow_mapped_task_facet(task_instance: TaskInstance) -> dict[str, Any]:
    # check for -1 comes from SmartSensor compatibility with dynamic task mapping
    # this comes from Airflow code
    log.debug(
        "AirflowMappedTaskRunFacet is deprecated and will be removed. "
        "Use information from AirflowRunFacet instead."
    )
    if hasattr(task_instance, "map_index") and getattr(task_instance, "map_index") != -1:
        return {"airflow_mappedTask": AirflowMappedTaskRunFacet.from_task_instance(task_instance)}
    return {}


def get_user_provided_run_facets(ti: TaskInstance, ti_state: TaskInstanceState) -> dict[str, RunFacet]:
    custom_facets = {}

    # Append custom run facets by executing the custom_run_facet functions.
    for custom_facet_func in conf.custom_run_facets():
        try:
            func: Callable[[TaskInstance, TaskInstanceState], dict[str, RunFacet]] | None = (
                try_import_from_string(custom_facet_func)
            )
            if not func:
                log.warning(
                    "OpenLineage is unable to import custom facet function `%s`; will ignore it.",
                    custom_facet_func,
                )
                continue
            facets: dict[str, RunFacet] | None = func(ti, ti_state)
            if facets and isinstance(facets, dict):
                duplicate_facet_keys = [facet_key for facet_key in facets if facet_key in custom_facets]
                if duplicate_facet_keys:
                    log.warning(
                        "Duplicate OpenLineage custom facets key(s) found: `%s` from function `%s`; "
                        "this will overwrite the previous value.",
                        ", ".join(duplicate_facet_keys),
                        custom_facet_func,
                    )
                log.debug(
                    "Adding OpenLineage custom facet with key(s): `%s` from function `%s`.",
                    tuple(facets),
                    custom_facet_func,
                )
                custom_facets.update(facets)
        except Exception as exc:
            log.warning(
                "Error processing custom facet function `%s`; will ignore it. Error was: %s: %s",
                custom_facet_func,
                type(exc).__name__,
                exc,
            )
    return custom_facets


def get_fully_qualified_class_name(operator: AnyOperator) -> str:
    if isinstance(operator, (SerializedMappedOperator, SerializedBaseOperator)):
        # as in airflow.api_connexion.schemas.common_schema.ClassReferenceSchema
        return operator._task_module + "." + operator.task_type
    op_class = get_operator_class(operator)
    return op_class.__module__ + "." + op_class.__name__


def is_operator_disabled(operator: AnyOperator) -> bool:
    return get_fully_qualified_class_name(operator) in conf.disabled_operators()


def is_selective_lineage_enabled(obj: DAG | SerializedDAG | AnyOperator) -> bool:
    """If selective enable is active check if DAG or Task is enabled to emit events."""
    if not conf.selective_enable():
        return True
    if isinstance(obj, (DAG, SerializedDAG)):
        return is_dag_lineage_enabled(obj)
    if isinstance(obj, (BaseOperator, MappedOperator)):
        return is_task_lineage_enabled(obj)
    raise TypeError("is_selective_lineage_enabled can only be used on DAG or Operator objects")


if not AIRFLOW_V_3_0_PLUS:

    @provide_session
    def is_ti_rescheduled_already(ti: TaskInstance, session=NEW_SESSION):
        from sqlalchemy import exists, select

        if not isinstance(ti.task, BaseSensorOperator):
            return False

        if not ti.task.reschedule:
            return False
        if AIRFLOW_V_3_0_PLUS:
            return (
                session.scalar(
                    select(
                        exists().where(
                            TaskReschedule.ti_id == ti.id, TaskReschedule.try_number == ti.try_number
                        )
                    )
                )
                is True
            )
        return (
            session.scalar(
                select(
                    exists().where(
                        TaskReschedule.dag_id == ti.dag_id,
                        TaskReschedule.task_id == ti.task_id,
                        TaskReschedule.run_id == ti.run_id,
                        TaskReschedule.map_index == ti.map_index,
                        TaskReschedule.try_number == ti.try_number,
                    )
                )
            )
            is True
        )


class InfoJsonEncodable(dict):
    """
    Airflow objects might not be json-encodable overall.

    The class provides additional attributes to control
    what and how is encoded:

    * renames: a dictionary of attribute name changes
    * | casts: a dictionary consisting of attribute names
      | and corresponding methods that should change
      | object value
    * includes: list of attributes to be included in encoding
    * excludes: list of attributes to be excluded from encoding

    Don't use both includes and excludes.
    """

    renames: dict[str, str] = {}
    casts: dict[str, Any] = {}
    includes: list[str] = []
    excludes: list[str] = []

    def __init__(self, obj):
        self.obj = obj
        self._fields = []

        self._cast_fields()
        self._rename_fields()
        self._include_fields()
        dict.__init__(
            self,
            **{field: InfoJsonEncodable._cast_basic_types(getattr(self, field)) for field in self._fields},
        )
        del self.obj

    @staticmethod
    def _cast_basic_types(value):
        if isinstance(value, datetime.datetime):
            return value.isoformat()
        if isinstance(value, datetime.timedelta):
            return f"{value.total_seconds()} seconds"
        if isinstance(value, (set, list, tuple)):
            return str(list(value))
        return value

    def _rename_fields(self):
        for field, renamed in self.renames.items():
            if hasattr(self.obj, field):
                setattr(self, renamed, getattr(self.obj, field))
                self._fields.append(renamed)

    def _cast_fields(self):
        for field, func in self.casts.items():
            setattr(self, field, func(self.obj))
            self._fields.append(field)

    def _include_fields(self):
        if self.includes and self.excludes:
            raise ValueError("Don't use both includes and excludes.")
        if self.includes:
            for field in self.includes:
                if field not in self._fields and hasattr(self.obj, field):
                    setattr(self, field, getattr(self.obj, field))
                    self._fields.append(field)
        else:
            if hasattr(self.obj, "__dict__"):
                obj_fields = self.obj.__dict__
            elif attrs.has(self.obj.__class__):  # e.g. attrs.define class with slots=True has no __dict__
                obj_fields = {
                    field.name: getattr(self.obj, field.name) for field in attrs.fields(self.obj.__class__)
                }
            else:
                raise ValueError(
                    "Cannot iterate over fields: "
                    f"The object of type {type(self.obj).__name__} neither has a __dict__ attribute "
                    "nor is defined as an attrs class."
                )
            for field, val in obj_fields.items():
                if field not in self._fields and field not in self.excludes and field not in self.renames:
                    setattr(self, field, val)
                    self._fields.append(field)


class DagInfo(InfoJsonEncodable):
    """Defines encoding DAG object to JSON."""

    includes = [
        "dag_id",
        "description",
        "fileloc",
        "owner",
        "owner_links",
        "schedule_interval",  # For Airflow 2 only -> AF3 has timetable_summary
        "start_date",
        "tags",
    ]
    casts = {
        "timetable": lambda dag: DagInfo.serialize_timetable(dag),
        "timetable_summary": lambda dag: DagInfo.timetable_summary(dag),
    }
    renames = {"_dag_id": "dag_id"}

    @classmethod
    def timetable_summary(cls, dag: DAG | SerializedDAG) -> str | None:
        """Extract summary from timetable if missing a ``timetable_summary`` property."""
        if summary := getattr(dag, "timetable_summary", None):
            return summary
        if (timetable := getattr(dag, "timetable", None)) is None:
            return None
        if summary := getattr(timetable, "summary", None):
            return summary
        with suppress(ImportError):
            from airflow.serialization.encoders import coerce_to_core_timetable

            return coerce_to_core_timetable(timetable).summary
        return None

    @classmethod
    def serialize_timetable(cls, dag: DAG | SerializedDAG) -> dict[str, Any]:
        # This is enough for Airflow 2.10+ and has all the information needed
        try:
            serialized = dag.timetable.serialize() or {}  # type: ignore[union-attr]
        except AttributeError:
            from airflow.serialization.encoders import encode_timetable

            serialized = encode_timetable(dag.timetable)["__var"]

        # In Airflow 2.9 when using Dataset scheduling we do not receive datasets in serialized timetable
        # Also for DatasetOrTimeSchedule, we only receive timetable without dataset_condition
        if hasattr(dag, "dataset_triggers") and "dataset_condition" not in serialized:
            try:
                # Make sure we are in Airflow version where these are importable
                from airflow.datasets import BaseDatasetEventInput, DatasetAll, DatasetAny
            except ImportError:
                log.warning("OpenLineage could not serialize full dag's timetable for dag `%s`.", dag.dag_id)
                return serialized

            def _serialize_ds(ds: BaseDatasetEventInput) -> dict[str, Any]:
                if isinstance(ds, (DatasetAny, DatasetAll)):
                    return {
                        "__type": "dataset_all" if isinstance(ds, DatasetAll) else "dataset_any",
                        "objects": [_serialize_ds(child) for child in ds.objects],
                    }
                return {"__type": "dataset", "uri": ds.uri, "extra": ds.extra}

            if isinstance(dag.dataset_triggers, BaseDatasetEventInput):
                serialized["dataset_condition"] = _serialize_ds(dag.dataset_triggers)
            elif isinstance(dag.dataset_triggers, list) and len(dag.dataset_triggers):
                serialized["dataset_condition"] = {
                    "__type": "dataset_all",
                    "objects": [_serialize_ds(trigger) for trigger in dag.dataset_triggers],
                }
        return serialized


class DagRunInfo(InfoJsonEncodable):
    """Defines encoding DagRun object to JSON."""

    includes = [
        "conf",
        "dag_id",
        "data_interval_start",
        "data_interval_end",
        "external_trigger",  # Removed in Airflow 3, use run_type instead
        "execution_date",  # Airflow 2
        "logical_date",  # Airflow 3
        "run_after",  # Airflow 3
        "run_id",
        "run_type",
        "start_date",
        "end_date",
    ]

    casts = {
        "duration": lambda dagrun: DagRunInfo.duration(dagrun),
        "dag_bundle_name": lambda dagrun: DagRunInfo.dag_version_info(dagrun, "bundle_name"),
        "dag_bundle_version": lambda dagrun: DagRunInfo.dag_version_info(dagrun, "bundle_version"),
        "dag_version_id": lambda dagrun: DagRunInfo.dag_version_info(dagrun, "version_id"),
        "dag_version_number": lambda dagrun: DagRunInfo.dag_version_info(dagrun, "version_number"),
    }

    @classmethod
    def duration(cls, dagrun: DagRun) -> float | None:
        if not getattr(dagrun, "end_date", None) or not isinstance(dagrun.end_date, datetime.datetime):
            return None
        if not getattr(dagrun, "start_date", None) or not isinstance(dagrun.start_date, datetime.datetime):
            return None
        return (dagrun.end_date - dagrun.start_date).total_seconds()

    @classmethod
    def dag_version_info(cls, dagrun: DagRun, key: str) -> str | int | None:
        # AF2 DagRun and AF3 DagRun SDK model (on worker) do not have this information
        if not getattr(dagrun, "dag_versions", []):
            return None
        current_version = dagrun.dag_versions[-1]
        if key == "bundle_name":
            return current_version.bundle_name
        if key == "bundle_version":
            return current_version.bundle_version
        if key == "version_id":
            return str(current_version.id)
        if key == "version_number":
            return current_version.version_number
        raise ValueError(f"Unsupported key: {key}`")


class TaskInstanceInfo(InfoJsonEncodable):
    """Defines encoding TaskInstance object to JSON."""

    includes = ["duration", "try_number", "pool", "queued_dttm", "log_url"]
    casts = {
        "log_url": lambda ti: getattr(ti, "log_url", None),
        "map_index": lambda ti: ti.map_index if getattr(ti, "map_index", -1) != -1 else None,
        "dag_bundle_version": lambda ti: (
            ti.bundle_instance.version if hasattr(ti, "bundle_instance") else None
        ),
        "dag_bundle_name": lambda ti: ti.bundle_instance.name if hasattr(ti, "bundle_instance") else None,
    }


class AssetInfo(InfoJsonEncodable):
    """Defines encoding Airflow Asset object to JSON."""

    includes = ["uri", "extra"]


class TaskInfo(InfoJsonEncodable):
    """Defines encoding BaseOperator/AbstractOperator object to JSON."""

    renames = {
        "_BaseOperator__from_mapped": "mapped",
        "_downstream_task_ids": "downstream_task_ids",
        "_upstream_task_ids": "upstream_task_ids",
        "_is_setup": "is_setup",
        "_is_teardown": "is_teardown",
    }
    includes = [
        "deferrable",
        "depends_on_past",
        "downstream_task_ids",
        "execution_timeout",
        "executor_config",
        "ignore_first_depends_on_past",
        "max_active_tis_per_dag",
        "max_active_tis_per_dagrun",
        "max_retry_delay",
        "multiple_outputs",
        "owner",
        "priority_weight",
        "queue",
        "retries",
        "retry_exponential_backoff",
        "run_as_user",
        "sla",
        "task_id",
        "trigger_rule",
        "upstream_task_ids",
        "wait_for_downstream",
        "wait_for_past_depends_before_skipping",
        # Operator-specific useful attributes
        "trigger_dag_id",  # TriggerDagRunOperator
        "trigger_run_id",  # TriggerDagRunOperator
        "external_dag_id",  # ExternalTaskSensor and ExternalTaskMarker (if run, as it's EmptyOperator)
        "external_task_id",  # ExternalTaskSensor and ExternalTaskMarker (if run, as it's EmptyOperator)
        "external_task_ids",  # ExternalTaskSensor
        "external_task_group_id",  # ExternalTaskSensor
        "external_dates_filter",  # ExternalTaskSensor
        "logical_date",  # AF 3 ExternalTaskMarker (if run, as it's EmptyOperator)
        "execution_date",  # AF 2 ExternalTaskMarker (if run, as it's EmptyOperator)
        "database",  # BaseSQlOperator
        "parameters",  # SQLCheckOperator, SQLValueCheckOperator and BranchSQLOperator
        "column_mapping",  # SQLColumnCheckOperator
        "pass_value",  # SQLValueCheckOperator
        "tol",  # SQLValueCheckOperator
        "metrics_thresholds",  # SQLIntervalCheckOperator
        "ratio_formula",  # SQLIntervalCheckOperator
        "ignore_zero",  # SQLIntervalCheckOperator
        "min_threshold",  # SQLThresholdCheckOperator
        "max_threshold",  # SQLThresholdCheckOperator
        "follow_task_ids_if_true",  # BranchSQLOperator
        "follow_task_ids_if_false",  # BranchSQLOperator
        "follow_branch",  # BranchSQLOperator
        "preoperator",  # SQLInsertRowsOperator
        "postoperator",  # SQLInsertRowsOperator
        "table_name_with_schema",  # SQLInsertRowsOperator
        "column_names",  # SQLInsertRowsOperator
    ]
    casts = {
        "operator_class": lambda task: task.task_type,
        "operator_class_path": lambda task: get_fully_qualified_class_name(task),
        "task_group": lambda task: (
            TaskGroupInfo(task.task_group)
            if hasattr(task, "task_group") and getattr(task.task_group, "_group_id", None)
            else None
        ),
        "inlets": lambda task: [AssetInfo(i) for i in task.inlets if isinstance(i, Asset)],
        "outlets": lambda task: [AssetInfo(o) for o in task.outlets if isinstance(o, Asset)],
        "operator_provider_version": lambda task: get_operator_provider_version(task),
    }


class TaskInfoComplete(TaskInfo):
    """Defines encoding BaseOperator/AbstractOperator object to JSON used when user enables full task info."""

    includes = []
    excludes = [
        "_BaseOperator__instantiated",
        "_dag",
        "_hook",
        "_log",
        "_outlets",
        "_inlets",
        "_lock_for_execution",
        "handler",
        "params",
        "python_callable",
        "retry_delay",
    ]


class TaskGroupInfo(InfoJsonEncodable):
    """Defines encoding TaskGroup object to JSON."""

    renames = {
        "_group_id": "group_id",
    }
    includes = [
        "downstream_group_ids",
        "downstream_task_ids",
        "prefix_group_id",
        "tooltip",
        "upstream_group_ids",
        "upstream_task_ids",
    ]


def get_airflow_dag_run_facet(dag_run: DagRun) -> dict[str, RunFacet]:
    if not dag_run.dag:
        return {}
    return {
        "airflowDagRun": AirflowDagRunFacet(
            dag=DagInfo(dag_run.dag),
            dagRun=DagRunInfo(dag_run),
        )
    }


@conf.cache
def _get_all_packages_installed() -> dict[str, str]:
    """
    Retrieve a dictionary of all installed packages and their versions.

    This operation involves scanning the system's installed packages, which can be a heavy operation.
    It is recommended to cache the result to avoid repeated, expensive lookups.
    """
    return {dist.metadata["Name"]: dist.version for dist in metadata.distributions()}


def get_processing_engine_facet() -> dict[str, processing_engine_run.ProcessingEngineRunFacet]:
    from openlineage.client.facet_v2 import processing_engine_run

    return {
        "processing_engine": processing_engine_run.ProcessingEngineRunFacet(
            version=AIRFLOW_VERSION,
            name="Airflow",
            openlineageAdapterVersion=OPENLINEAGE_PROVIDER_VERSION,
        )
    }


def get_airflow_debug_facet() -> dict[str, AirflowDebugRunFacet]:
    if not conf.debug_mode():
        return {}
    log.warning("OpenLineage debug_mode is enabled. Be aware that this may log and emit extensive details.")
    return {
        "debug": AirflowDebugRunFacet(
            packages=_get_all_packages_installed(),
        )
    }


def get_airflow_run_facet(
    dag_run: DagRun,
    dag: DAG | SerializedDAG,
    task_instance: TaskInstance,
    task: BaseOperator,
    task_uuid: str,
) -> dict[str, AirflowRunFacet]:
    return {
        "airflow": AirflowRunFacet(
            dag=DagInfo(dag),
            dagRun=DagRunInfo(dag_run),
            taskInstance=TaskInstanceInfo(task_instance),
            task=TaskInfoComplete(task) if conf.include_full_task_info() else TaskInfo(task),
            taskUuid=task_uuid,
        )
    }


def get_airflow_job_facet(dag_run: DagRun) -> dict[str, AirflowJobFacet]:
    if not dag_run.dag:
        return {}
    return {
        "airflow": AirflowJobFacet(
            taskTree={},  # caused OOM errors, to be removed, see #41587
            taskGroups=_get_task_groups_details(dag_run.dag),
            tasks=_get_tasks_details(dag_run.dag),
        )
    }


def get_airflow_state_run_facet(
    dag_id: str, run_id: str, task_ids: list[str], dag_run_state: DagRunState
) -> dict[str, AirflowStateRunFacet]:
    tis = DagRun.fetch_task_instances(dag_id=dag_id, run_id=run_id, task_ids=task_ids)

    def get_task_duration(ti):
        if ti.duration is not None:
            return ti.duration
        if ti.end_date is not None and ti.start_date is not None:
            return (ti.end_date - ti.start_date).total_seconds()
        # Fallback to 0.0 for tasks with missing timestamps (e.g., skipped/terminated tasks)
        return 0.0

    return {
        "airflowState": AirflowStateRunFacet(
            dagRunState=dag_run_state,
            tasksState={ti.task_id: ti.state for ti in tis},
            tasksDuration={ti.task_id: get_task_duration(ti) for ti in tis},
        )
    }


def _get_tasks_details(dag: DAG | SerializedDAG) -> dict:
    tasks = {
        single_task.task_id: {
            "operator": get_fully_qualified_class_name(single_task),
            "task_group": single_task.task_group.group_id if single_task.task_group else None,
            "emits_ol_events": _emits_ol_events(single_task),
            "ui_color": single_task.ui_color,
            "ui_fgcolor": single_task.ui_fgcolor,
            "ui_label": single_task.label,
            "is_setup": single_task.is_setup,
            "is_teardown": single_task.is_teardown,
            "downstream_task_ids": sorted(single_task.downstream_task_ids),
        }
        for single_task in sorted(dag.tasks, key=lambda x: x.task_id)
    }

    return tasks


def _get_task_groups_details(dag: DAG | SerializedDAG) -> dict:
    return {
        tg_id: {
            "parent_group": tg.parent_group.group_id,
            "ui_color": tg.ui_color,
            "ui_fgcolor": tg.ui_fgcolor,
            "ui_label": tg.label,
        }
        for tg_id, tg in dag.task_group_dict.items()
    }


def _emits_ol_events(task: AnyOperator) -> bool:
    config_selective_enabled = is_selective_lineage_enabled(task)
    config_disabled_for_operators = is_operator_disabled(task)

    is_task_schedulable_method = getattr(TaskInstance, "is_task_schedulable", None)  # Added in 3.2.0 #56039
    if is_task_schedulable_method and callable(is_task_schedulable_method):
        is_skipped_as_empty_operator = not is_task_schedulable_method(task)
    else:
        # For older Airflow versions, re-create Airflow core internal logic as
        # empty operators without callbacks/outlets are skipped for optimization by Airflow
        # in airflow.models.taskinstance.TaskInstance._schedule_downstream_tasks or
        # airflow.models.dagrun.DagRun.schedule_tis, depending on Airflow version
        is_skipped_as_empty_operator = all(
            (
                task.inherits_from_empty_operator,
                not getattr(task, "on_execute_callback", None),
                not getattr(task, "on_success_callback", None),
                not task.outlets,
                not (task.inlets and get_base_airflow_version_tuple() >= (3, 0, 2)),  # Added in 3.0.2 #50773
                not (
                    getattr(task, "has_on_execute_callback", None)  # Added in 3.1.0 #54569
                    and get_base_airflow_version_tuple() >= (3, 1, 0)
                ),
                not (
                    getattr(task, "has_on_success_callback", None)  # Added in 3.1.0 #54569
                    and get_base_airflow_version_tuple() >= (3, 1, 0)
                ),
            )
        )

    emits_ol_events = all(
        (
            config_selective_enabled,
            not config_disabled_for_operators,
            not is_skipped_as_empty_operator,
        )
    )
    return emits_ol_events


def get_unknown_source_attribute_run_facet(task: BaseOperator, name: str | None = None):
    if not name:
        name = get_operator_class(task).__name__
    log.debug(
        "UnknownOperatorAttributeRunFacet is deprecated and will be removed. "
        "Use information from AirflowRunFacet instead."
    )
    return {
        "unknownSourceAttribute": attrs.asdict(
            UnknownOperatorAttributeRunFacet(
                unknownItems=[
                    UnknownOperatorInstance(
                        name=name,
                        properties=TaskInfo(task),
                    )
                ]
            )
        )
    }


class OpenLineageRedactor(SecretsMasker):
    """
    This class redacts sensitive data similar to SecretsMasker in Airflow logs.

    The difference is that our default max recursion depth is way higher - due to
    the structure of OL events we need more depth.
    Additionally, we allow data structures to specify data that needs not to be
    redacted by specifying _skip_redact list by deriving RedactMixin.
    """

    @classmethod
    def from_masker(cls, other: SecretsMasker) -> OpenLineageRedactor:
        instance = cls()
        instance.patterns = other.patterns
        instance.replacer = other.replacer
        for attr in ["sensitive_variables_fields", "min_length_to_mask", "secret_mask_adapter"]:
            if hasattr(other, attr):
                setattr(instance, attr, getattr(other, attr))
        return instance

    def _should_hide_value_for_key(self, name):
        """Compatibility helper for should_hide_value_for_key across Airflow versions."""
        try:
            return self.should_hide_value_for_key(name)
        except AttributeError:
            # fallback to module level function
            return should_hide_value_for_key(name)

    def _redact(self, item: Redactable, name: str | None, depth: int, max_depth: int, **kwargs) -> Redacted:  # type: ignore[override]
        if AIRFLOW_V_3_0_PLUS:
            # Keep compatibility for Airflow 2.x, remove when Airflow 3.0 is the minimum version
            class AirflowContextDeprecationWarning(UserWarning):
                pass
        else:
            from airflow.utils.context import (  # type: ignore[attr-defined,no-redef]
                AirflowContextDeprecationWarning,
            )

        if depth > max_depth:
            return item
        try:
            # It's impossible to check the type of variable in a dict without accessing it, and
            # this already causes warning - so suppress it
            with suppress(AirflowContextDeprecationWarning):
                if type(item).__name__ == "Proxy":
                    # Those are deprecated values in _DEPRECATION_REPLACEMENTS
                    # in airflow.utils.context.Context
                    return "<<non-redactable: Proxy>>"
                if name and self._should_hide_value_for_key(name):
                    return self._redact_all(item, depth, max_depth)
                if attrs.has(type(item)):
                    # TODO: FIXME when mypy gets compatible with new attrs
                    for dict_key, subval in attrs.asdict(
                        item,  # type: ignore[arg-type]
                        recurse=False,
                    ).items():
                        if _is_name_redactable(dict_key, item):
                            setattr(
                                item,
                                dict_key,
                                self._redact(
                                    subval,
                                    name=dict_key,
                                    depth=(depth + 1),
                                    max_depth=max_depth,
                                ),
                            )
                    return item
                if is_json_serializable(item) and hasattr(item, "__dict__"):
                    for dict_key, subval in item.__dict__.items():
                        if type(subval).__name__ == "Proxy":
                            return "<<non-redactable: Proxy>>"
                        if _is_name_redactable(dict_key, item):
                            setattr(
                                item,
                                dict_key,
                                self._redact(
                                    subval,
                                    name=dict_key,
                                    depth=(depth + 1),
                                    max_depth=max_depth,
                                ),
                            )
                    return item
                return super()._redact(item, name, depth, max_depth, **kwargs)
        except Exception as exc:
            log.warning("Unable to redact %r. Error was: %s: %s", item, type(exc).__name__, exc)
        return item


def is_json_serializable(item):
    try:
        json.dumps(item)
        return True
    except (TypeError, ValueError):
        return False


def _is_name_redactable(name, redacted):
    if not issubclass(redacted.__class__, RedactMixin):
        return not name.startswith("_")
    return name not in redacted.skip_redact


def print_warning(log):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception:
                log.warning(
                    "OpenLineage event emission failed. "
                    "Exception below is being caught but it's printed for visibility. "
                    "This has no impact on actual task execution status.",
                    exc_info=True,
                )

        return wrapper

    return decorator


def get_filtered_unknown_operator_keys(operator: BaseOperator) -> dict:
    not_required_keys = {"dag", "task_group"}
    return {attr: value for attr, value in operator.__dict__.items() if attr not in not_required_keys}


def should_use_external_connection(hook) -> bool:
    # If we're at Airflow 2.10, the execution is process-isolated, so we can safely run those again.
    return True


def translate_airflow_asset(asset: Asset, lineage_context) -> OpenLineageDataset | None:
    """
    Convert an Asset with an AIP-60 compliant URI to an OpenLineageDataset.

    This function returns None if no URI normalizer is defined, no asset converter is found or
    some core Airflow changes are missing and ImportError is raised.
    """
    if AIRFLOW_V_3_0_PLUS:
        from airflow.sdk.definitions.asset import _get_normalized_scheme
    else:
        try:
            from airflow.datasets import _get_normalized_scheme  # type: ignore[no-redef]
        except ImportError:
            return None

    try:
        from airflow.providers_manager import ProvidersManager

        ol_converters = getattr(ProvidersManager(), "asset_to_openlineage_converters", None)
        if not ol_converters:
            ol_converters = ProvidersManager().dataset_to_openlineage_converters  # type: ignore[attr-defined]

        normalized_uri = asset.normalized_uri
    except (ImportError, AttributeError):
        return None

    if normalized_uri is None:
        return None

    if not (normalized_scheme := _get_normalized_scheme(normalized_uri)):
        return None

    if (airflow_to_ol_converter := ol_converters.get(normalized_scheme)) is None:
        return None

    return airflow_to_ol_converter(Asset(uri=normalized_uri, extra=asset.extra), lineage_context)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any
from unittest import mock

import pytest
from attrs import Factory, define, field
from openlineage.client.event_v2 import Dataset
from openlineage.client.facet_v2 import BaseFacet, JobFacet, parent_run, sql_job

from airflow.models.taskinstance import TaskInstanceState
from airflow.providers.common.compat.sdk import BaseOperator
from airflow.providers.openlineage.extractors.base import (
    BaseExtractor,
    DefaultExtractor,
    OperatorLineage,
)
from airflow.providers.openlineage.extractors.manager import ExtractorManager
from airflow.providers.openlineage.extractors.python import PythonExtractor

from tests_common.test_utils.compat import PythonOperator

if TYPE_CHECKING:
    from openlineage.client.facet_v2 import RunFacet


INPUTS = [Dataset(namespace="database://host:port", name="inputtable")]
OUTPUTS = [Dataset(namespace="database://host:port", name="inputtable")]
RUN_FACETS: dict[str, RunFacet] = {
    "parent": parent_run.ParentRunFacet(
        run=parent_run.Run(runId="3bb703d1-09c1-4a42-8da5-35a0b3216072"),
        job=parent_run.Job(namespace="namespace", name="parentjob"),
    )
}
JOB_FACETS: dict[str, JobFacet] = {"sql": sql_job.SQLJobFacet(query="SELECT * FROM inputtable")}


@define
class CompleteRunFacet(JobFacet):
    finished: bool = field(default=False)


@define
class FailRunFacet(JobFacet):
    failed: bool = field(default=False)


FINISHED_FACETS: dict[str, JobFacet] = {"complete": CompleteRunFacet(True)}
FAILED_FACETS: dict[str, JobFacet] = {"failure": FailRunFacet(True)}


class ExtractorWithoutExecuteOnFailure(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls):
        return ["SimpleCustomOperator"]

    def _execute_extraction(self) -> OperatorLineage | None:
        return OperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=JOB_FACETS,
        )

    def extract_on_complete(self, task_instance) -> OperatorLineage | None:
        return OperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=FINISHED_FACETS,
        )


class ExtractorWithExecuteExtractionOnly(BaseExtractor):
    @classmethod
    def get_operator_classnames(cls):
        return ["AnotherOperator"]

    def _execute_extraction(self) -> OperatorLineage | None:
        return OperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=JOB_FACETS,
        )


class SimpleCustomOperator(BaseOperator):
    def execute(self, context) -> Any:
        pass

    def get_openlineage_facets_on_start(self) -> OperatorLineage:
        return OperatorLineage()

    def get_openlineage_facets_on_complete(self, task_instance) -> OperatorLineage:
        return OperatorLineage()

    def get_openlineage_facets_on_failure(self, task_instance) -> OperatorLineage:
        return OperatorLineage()


class OperatorWithoutFailure(BaseOperator):
    def execute(self, context) -> Any:
        pass

    def get_openlineage_facets_on_start(self) -> OperatorLineage:
        return OperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=JOB_FACETS,
        )

    def get_openlineage_facets_on_complete(self, task_instance) -> OperatorLineage:
        return OperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=FINISHED_FACETS,
        )


class OperatorWithAllOlMethods(BaseOperator):
    def execute(self, context) -> Any:
        pass

    def get_openlineage_facets_on_start(self) -> OperatorLineage:
        return OperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=JOB_FACETS,
        )

    def get_openlineage_facets_on_complete(self, task_instance) -> OperatorLineage:
        return OperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=FINISHED_FACETS,
        )

    def get_openlineage_facets_on_failure(self, task_instance) -> OperatorLineage:
        return OperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=FAILED_FACETS,
        )


class OperatorWithoutComplete(BaseOperator):
    def execute(self, context) -> Any:
        pass

    def get_openlineage_facets_on_start(self) -> OperatorLineage:
        return OperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=JOB_FACETS,
        )


class OperatorWithoutStart(BaseOperator):
    def execute(self, context) -> Any:
        pass

    def get_openlineage_facets_on_complete(self, task_instance) -> OperatorLineage:
        return OperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=FINISHED_FACETS,
        )


class OperatorDifferentOperatorLineageClass(BaseOperator):
    def execute(self, context) -> Any:
        pass

    def get_openlineage_facets_on_start(self):
        @define
        class DifferentOperatorLineage:
            name: str = ""
            inputs: list[Dataset] = Factory(list)
            outputs: list[Dataset] = Factory(list)
            run_facets: dict[str, BaseFacet] = Factory(dict)
            job_facets: dict[str, BaseFacet] = Factory(dict)
            some_other_param: dict = Factory(dict)

        return DifferentOperatorLineage(
            name="unused",
            inputs=INPUTS,
            outputs=OUTPUTS,
            run_facets=RUN_FACETS,
            job_facets=JOB_FACETS,
            some_other_param={"asdf": "fdsa"},
        )


class OperatorWrongOperatorLineageClass(BaseOperator):
    def execute(self, context) -> Any:
        pass

    def get_openlineage_facets_on_start(self):
        @define
        class WrongOperatorLineage:
            inputs: list[Dataset] = Factory(list)
            outputs: list[Dataset] = Factory(list)
            some_other_param: dict = Factory(dict)

        return WrongOperatorLineage(
            inputs=INPUTS,
            outputs=OUTPUTS,
            some_other_param={"asdf": "fdsa"},
        )


class BrokenOperator(BaseOperator):
    get_openlineage_facets: list[BaseFacet] = []

    def execute(self, context) -> Any:
        pass


def test_default_extraction():
    extractor = ExtractorManager().get_extractor_class(OperatorWithoutFailure)
    assert extractor is DefaultExtractor

    metadata = extractor(OperatorWithoutFailure(task_id="test")).extract()

    task_instance = mock.MagicMock()

    metadata_on_complete = extractor(OperatorWithoutFailure(task_id="test")).extract_on_complete(
        task_instance=task_instance
    )

    assert metadata == OperatorLineage(
        inputs=INPUTS,
        outputs=OUTPUTS,
        run_facets=RUN_FACETS,
        job_facets=JOB_FACETS,
    )

    assert metadata_on_complete == OperatorLineage(
        inputs=INPUTS,
        outputs=OUTPUTS,
        run_facets=RUN_FACETS,
        job_facets=FINISHED_FACETS,
    )


def test_extraction_without_on_complete():
    extractor = ExtractorManager().get_extractor_class(OperatorWithoutComplete)
    assert extractor is DefaultExtractor

    metadata = extractor(OperatorWithoutComplete(task_id="test")).extract()

    task_instance = mock.MagicMock()

    metadata_on_complete = extractor(OperatorWithoutComplete(task_id="test")).extract_on_complete(
        task_instance=task_instance
    )

    expected_task_metadata = OperatorLineage(
        inputs=INPUTS,
        outputs=OUTPUTS,
        run_facets=RUN_FACETS,
        job_facets=JOB_FACETS,
    )

    assert metadata == expected_task_metadata

    assert metadata_on_complete == expected_task_metadata


def test_extraction_without_on_start():
    extractor = ExtractorManager().get_extractor_class(OperatorWithoutStart)
    assert extractor is DefaultExtractor

    metadata = extractor(OperatorWithoutStart(task_id="test")).extract()

    task_instance = mock.MagicMock()

    metadata_on_complete = extractor(OperatorWithoutStart(task_id="test")).extract_on_complete(
        task_instance=task_instance
    )

    assert metadata == OperatorLineage()

    assert metadata_on_complete == OperatorLineage(
        inputs=INPUTS,
        outputs=OUTPUTS,
        run_facets=RUN_FACETS,
        job_facets=FINISHED_FACETS,
    )


@pytest.mark.parametrize(
    ("operator_class", "task_state", "expected_job_facets"),
    (
        (OperatorWithAllOlMethods, TaskInstanceState.FAILED, FAILED_FACETS),
        (OperatorWithAllOlMethods, TaskInstanceState.RUNNING, JOB_FACETS),
        (OperatorWithAllOlMethods, TaskInstanceState.SUCCESS, FINISHED_FACETS),
        (OperatorWithAllOlMethods, TaskInstanceState.UP_FOR_RETRY, FINISHED_FACETS),  # Should never happen
        (OperatorWithAllOlMethods, None, FINISHED_FACETS),  # Should never happen
        (OperatorWithoutFailure, TaskInstanceState.FAILED, FINISHED_FACETS),
        (OperatorWithoutFailure, TaskInstanceState.RUNNING, JOB_FACETS),
        (OperatorWithoutFailure, TaskInstanceState.SUCCESS, FINISHED_FACETS),
        (OperatorWithoutFailure, TaskInstanceState.UP_FOR_RETRY, FINISHED_FACETS),  # Should never happen
        (OperatorWithoutFailure, None, FINISHED_FACETS),  # Should never happen
        (OperatorWithoutStart, TaskInstanceState.FAILED, FINISHED_FACETS),
        (OperatorWithoutStart, TaskInstanceState.RUNNING, {}),
        (OperatorWithoutStart, TaskInstanceState.SUCCESS, FINISHED_FACETS),
        (OperatorWithoutStart, TaskInstanceState.UP_FOR_RETRY, FINISHED_FACETS),  # Should never happen
        (OperatorWithoutStart, None, FINISHED_FACETS),  # Should never happen
        (OperatorWithoutComplete, TaskInstanceState.FAILED, JOB_FACETS),
        (OperatorWithoutComplete, TaskInstanceState.RUNNING, JOB_FACETS),
        (OperatorWithoutComplete, TaskInstanceState.SUCCESS, JOB_FACETS),
        (OperatorWithoutComplete, TaskInstanceState.UP_FOR_RETRY, JOB_FACETS),  # Should never happen
        (OperatorWithoutComplete, None, JOB_FACETS),  # Should never happen
    ),
)
def test_extractor_manager_calls_appropriate_extractor_method(
    operator_class, task_state, expected_job_facets
):
    extractor_manager = ExtractorManager()

    ti = mock.MagicMock()

    metadata = extractor_manager.extract_metadata(
        dagrun=mock.MagicMock(run_id="dagrun_run_id"),
        task=operator_class(task_id="task_id"),
        task_instance_state=task_state,
        task_instance=ti,
    )

    assert metadata.job_facets == expected_job_facets
    if not expected_job_facets:  # Empty OperatorLineage() is expected
        assert not metadata.inputs
        assert not metadata.outputs
        assert not metadata.run_facets
    else:
        assert metadata.inputs == INPUTS
        assert metadata.outputs == OUTPUTS
        assert metadata.run_facets == RUN_FACETS


@mock.patch("airflow.providers.openlineage.conf.custom_extractors")
def test_extractors_env_var(custom_extractors):
    custom_extractors.return_value = {
        "unit.openlineage.extractors.test_base.ExtractorWithoutExecuteOnFailure"
    }
    extractor = ExtractorManager().get_extractor_class(SimpleCustomOperator(task_id="example"))
    assert extractor is ExtractorWithoutExecuteOnFailure


def test_extractor_without_extract_on_failure_calls_extract_on_complete():
    extractor = ExtractorWithoutExecuteOnFailure(SimpleCustomOperator(task_id="example"))
    result = extractor.extract_on_failure(None)
    assert result == OperatorLineage(
        inputs=INPUTS,
        outputs=OUTPUTS,
        run_facets=RUN_FACETS,
        job_facets=FINISHED_FACETS,
    )


def test_extractor_without_extract_on_complete_and_failure_always_calls_extract():
    extractor = ExtractorWithExecuteExtractionOnly(SimpleCustomOperator(task_id="example"))
    expected_result = OperatorLineage(
        inputs=INPUTS,
        outputs=OUTPUTS,
        run_facets=RUN_FACETS,
        job_facets=JOB_FACETS,
    )
    result = extractor.extract_on_failure(None)
    assert result == expected_result
    result = extractor.extract_on_complete(None)
    assert result == expected_result
    result = extractor.extract()
    assert result == expected_result


def test_does_not_use_default_extractor_when_not_a_method():
    extractor_class = ExtractorManager().get_extractor_class(BrokenOperator(task_id="a"))
    assert extractor_class is None


def test_does_not_use_default_extractor_when_no_get_openlineage_facets():
    extractor_class = ExtractorManager().get_extractor_class(BaseOperator(task_id="b"))
    assert extractor_class is None


def test_does_not_use_default_extractor_when_explicit_extractor():
    extractor_class = ExtractorManager().get_extractor_class(
        PythonOperator(task_id="c", python_callable=lambda: 7)
    )
    assert extractor_class is PythonExtractor


def test_default_extractor_uses_different_operatorlineage_class():
    operator = OperatorDifferentOperatorLineageClass(task_id="task_id")
    extractor_class = ExtractorManager().get_extractor_class(operator)
    assert extractor_class is DefaultExtractor
    extractor = extractor_class(operator)
    assert extractor.extract() == OperatorLineage(
        inputs=INPUTS,
        outputs=OUTPUTS,
        run_facets=RUN_FACETS,
        job_facets=JOB_FACETS,
    )


def test_default_extractor_uses_wrong_operatorlineage_class():
    operator = OperatorWrongOperatorLineageClass(task_id="task_id")
    # If extractor returns lineage class that can't be changed into OperatorLineage, just return
    # empty OperatorLineage
    assert ExtractorManager().extract_metadata(mock.MagicMock(), operator, None) == OperatorLineage()
